from __future__ import print_function

import socket
import logging
import time
import logging.handlers
import subprocess
import re
import sys
import shutil
import netifaces
import random
import pycurl
import argparse
import json
from io import StringIO
from .ping import ping, ping6
from .dnsbench import resolve
from .http import get_curl, request
import textwrap

class HelpFormatter(argparse.ArgumentDefaultsHelpFormatter):

    def _get_help_string(self, action):
        return super(HelpFormatter, self)._get_help_string(action) \
            if action.default else action.help

    def _split_lines(self, text, width):
        """Preserves new lines in option descriptions"""
        lines = []
        for text in text.splitlines():
            lines += textwrap.wrap(text, width)
        return lines

    def _fill_text(self, text, width, indent):
        """Preserves new lines in other descriptions"""
        kw = dict(width=width, initial_indent=indent, subsequent_indent=indent)
        return '\n'.join(textwrap.fill(t, **kw) for t in text.splitlines())

botname = socket.gethostname()

date_reg_exp = re.compile('\d{4}[-/]\d{2}[-/]\d{2}')

def _get_network_gateway(self):
  return netifaces.gateways()["default"][netifaces.AF_INET][0]

def load_configuration(config_path):
  if config_path.startswith("http://") or \
      config_path.startswith("ftp://") or \
      config_path.startswith("https://") or \
      config_path.startswith("file://"):
    return download_external_configuration(config_path)

  with open(config_path, "r") as f:
    return json.load(f)

def download_external_configuration(url):
  buffer = StringIO()
  curl, _ = get_curl(buffer, url)

  response_code = curl.getinfo(pycurl.HTTP_CODE)
  
  if response_code in (200, 0):
    try:
      return json.loads(buffer.getvalue())
    except ValueError:
      print("Unable to parse external configuration, error:")
      import traceback
      traceback.print_exc(file=sys.stderr)
      sys.stderr.flush()
      print("Ignoring external configuration")
    finally:
      curl.close()

  return {}

def is_rotate_log(log_file_path):
  try:
    log_file = open(log_file_path, 'r')
  except IOError:
    return False

  # Handling try-except-finally together.
  try:
    try:
      log_file.seek(0, 2)
      size = log_file.tell()
      log_file.seek(-min(size, 4096), 1)
      today = time.strftime("%Y-%m-%d")
      
      for line in reversed(log_file.read().split('\n')):
        if len(line.strip()):
          match_list = date_reg_exp.findall(line)
          if len(match_list):
            if match_list[0] != today:
              return ValueError(match_list[0])
            break

    except IOError:
      return False
  finally:
    log_file.close()

def rotate_logfile(handler, log_file):
  last_date = is_rotate_log(log_file)
  if last_date: 
    handler.doRollover()
    today = time.strftime("%Y-%m-%d")
    shutil.move("%s.%s" % (log_file, today),
                "%s.%s" % (log_file, last_date))
    # The simpler the better
    sp = subprocess.Popen("gzip %s.%s" % (log_file, last_date),
               stdout=subprocess.PIPE, stdin=subprocess.PIPE, shell=True)
    sp.communicate()

def create_logger(name, log_folder, verbose):
  new_logger = logging.getLogger(name)

  new_logger.setLevel(logging.DEBUG)
  log_file = '%s/network_bench.%s.log' % (log_folder, name)
  handler = logging.handlers.TimedRotatingFileHandler(
                     log_file, when="D",
                     backupCount=1000)

  rotate_logfile(handler, log_file)

  format = "%%(asctime)-16s;%s;%%(message)s" % botname
  handler.setFormatter(logging.Formatter(format))
  new_logger.addHandler(handler)

  if verbose:
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format))
    new_logger.addHandler(handler)
  return new_logger

def run_all(config_dict, log_folder, verbose):
  dns_logger = create_logger("dns", log_folder, verbose)
  name_dict = config_dict.get("dns", {})
  for name in name_dict:
    expected = name_dict[name].get("expected")
    dns_logger.info(';'.join(str(x) for x in resolve(name, expected)))

  ping_logger = create_logger("ping", log_folder, verbose)
  for host in config_dict.get("ping",[]):
    ping_logger.info(';'.join(str(x) for x in ping(host)))

  ping6_logger = create_logger("ping6", log_folder, verbose)
  for host in config_dict.get("ping6", []):
    ping6_logger.info(';'.join(str(x) for x in ping6(host)))

  http_logger = create_logger("http", log_folder, verbose)
  url_dict = config_dict.get("url", {})
  for url in url_dict:
    http_logger.info(';'.join(str(x) for x in request(url, url_dict[url])))

def main():
  parser = argparse.ArgumentParser(
        description="Run network benchmarch.",
        )
  _ = parser.add_argument
  _('-l', '--logdir', default=".",
                      help="Directory where the logs are going to be placed.")
  _('-c', '--conf', help="Path to the configuration json file.")
  _('-v', '--verbose', action='store_true', 
                       help="Show the results on stdout.")
  _('-d', '--delay', default=random.randint(0, 30), 
                     help="Delay before start to run," \
                          "as this script can be called on cron.")

  config = parser.parse_args()

  print("Downloading %s..." % config.conf.strip())
  config_dict = load_configuration(config.conf)

  print("Waiting %s before start..." % config.delay)
  time.sleep(float(config.delay))

  run_all(config_dict, 
          log_folder=config.logdir, 
          verbose=config.verbose)




from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise
from slapos.util import str2bytes
from zope.interface import implementer
import datetime


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def sense(self):
    """
      Check the certificate
    """

    certificate_file = self.getConfig('certificate')
    key_file = self.getConfig('key')

    try:
      certificate_expiration_days = int(
        self.getConfig('certificate-expiration-days', '15'))
    except ValueError:
      self.logger.error('ERROR certificate-expiration-days is wrong: %r' % (
        self.getConfig('certificate-expiration-days')))
      return

    try:
      with open(certificate_file, 'r') as fh:
        certificate = x509.load_pem_x509_certificate(
          str2bytes(fh.read()), default_backend())
    except Exception as e:
      self.logger.error(
        'ERROR Problem loading certificate %r, error: %s' % (
          certificate_file, e))
      return

    try:
      with open(key_file, 'r') as fh:
        key = serialization.load_pem_private_key(
          str2bytes(fh.read()), None, default_backend())
    except Exception as e:
      self.logger.error(
        'ERROR Problem loading key %r, error: %s' % (key_file, e))
      return

    if certificate.public_key().public_numbers() != \
       key.public_key().public_numbers():
      self.logger.error(
        'ERROR Certificate %r does not match key %r' % (
          certificate_file, key_file))
      return

    if certificate.not_valid_after - datetime.timedelta(
       days=certificate_expiration_days) < datetime.datetime.utcnow():
      self.logger.error(
       'ERROR Certificate %r will expire in less than %s days' % (
         certificate_file, certificate_expiration_days))
      return

    self.logger.info(
      'OK Certificate %r and key %r are ok' % (certificate_file, key_file))



##############################################################################
#
# Copyright (c) 2020 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from slapos.grid.promise import PromiseError
from slapos.test.promise.plugin import TestPromisePluginMixin
import datetime
import os
import shutil
import tempfile
import unittest
from slapos.util import bytes2str


class TestCheckCertificate(TestPromisePluginMixin):
  same_file = False
  promise_name = 'check-certificate.py'

  def setUp(self):
    super(TestCheckCertificate, self).setUp()
    self.tempdir = tempfile.mkdtemp()
    self.addCleanup(shutil.rmtree, self.tempdir)
    self.key_path = os.path.join(self.tempdir, 'key.pem')
    if self.same_file:
      self.certificate_path = self.key_path
    else:
      self.certificate_path = os.path.join(self.tempdir, 'certificate.pem')

  def createKey(self):
    key = rsa.generate_private_key(
      public_exponent=65537, key_size=2048, backend=default_backend())
    key_pem = key.private_bytes(
      encoding=serialization.Encoding.PEM,
      format=serialization.PrivateFormat.TraditionalOpenSSL,
      encryption_algorithm=serialization.NoEncryption()
    )
    return key, key_pem

  def createCertificate(self, key, days=30):
    subject = issuer = x509.Name([
      x509.NameAttribute(NameOID.COUNTRY_NAME, u"FR"),
      x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"Nord"),
      x509.NameAttribute(NameOID.LOCALITY_NAME, u"Lille"),
      x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"Nexedi"),
      x509.NameAttribute(NameOID.COMMON_NAME, u"Common"),
    ])
    certificate = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(
        datetime.datetime.utcnow() + datetime.timedelta(days)
    ).sign(key, hashes.SHA256(), default_backend())
    certificate_pem = certificate.public_bytes(
      encoding=serialization.Encoding.PEM)
    return certificate, certificate_pem

  def createKeyCertificate(self, days=30):
    key, key_pem = self.createKey()
    certificate, certificate_pem = self.createCertificate(key, days)
    with open(self.key_path, 'w') as fh:
      fh.write(bytes2str(key_pem))
    with open(self.certificate_path, 'a') as fh:
      fh.write(bytes2str(certificate_pem))

  def createKeyCertificateNotMatch(self):
    key, key_pem = self.createKey()
    another_key, another_key_pem = self.createKey()
    certificate, certificate_pem = self.createCertificate(key)
    with open(self.key_path, 'w') as fh:
      fh.write(bytes2str(another_key_pem))
    with open(self.certificate_path, 'a') as fh:
      fh.write(bytes2str(certificate_pem))

  def writePromise(self, d=None):
    if d is None:
      d = {}
    content_list = [
      "from slapos.promise.plugin.check_certificate import RunPromise"]
    content_list.append('extra_config_dict = {')
    for k, v in d.items():
      content_list.append("  '%s': '%s'," % (k, v))
    content_list.append('}')
    super(
      TestCheckCertificate, self).writePromise(
        self.promise_name, '\n'.join(content_list))

  def assertFailedMessage(self, result, message):
    self.assertEqual(result['result']['failed'], True)
    self.assertEqual(
      result['result']['message'],
      message)

  def assertPassedMessage(self, result, message):
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(
      result['result']['message'],
      message)

  def test(self):
    self.createKeyCertificate()
    self.writePromise({
      'certificate': self.certificate_path,
      'key': self.key_path
    })
    self.configureLauncher()
    self.launcher.run()
    self.assertPassedMessage(
      self.getPromiseResult(self.promise_name),
      "OK Certificate '%s' and key '%s' are ok" % (
        self.certificate_path, self.key_path)
    )

  def test_no_key(self):
    self.createKeyCertificate()
    nokey_path = os.path.join(self.tempdir, 'nokey.pem')
    self.writePromise({
      'certificate': self.certificate_path,
      'key': nokey_path,
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Problem loading key '%s', error: [Errno 2] No such file or "
      "directory: '%s'" % (nokey_path, nokey_path))

  def test_no_certificate(self):
    self.createKeyCertificate()
    nocertificate_path = os.path.join(self.tempdir, 'nocertificate.pem')
    self.writePromise({
      'certificate': nocertificate_path,
      'key': self.key_path
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Problem loading certificate '%s', error: [Errno 2] No such "
      "file or directory: '%s'" % (nocertificate_path, nocertificate_path))

  def test_does_not_match(self):
    self.createKeyCertificateNotMatch()
    self.writePromise({
      'certificate': self.certificate_path,
      'key': self.key_path
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Certificate '%s' does not match key '%s'" % (
        self.certificate_path, self.key_path)
    )

  def test_expires(self):
    self.createKeyCertificate(days=5)
    self.writePromise({
      'certificate': self.certificate_path,
      'key': self.key_path
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Certificate '%s' will expire in less than 15 days" % (
        self.certificate_path,)
    )

  def test_expires_custom(self):
    self.createKeyCertificate(days=19)
    self.writePromise({
      'certificate': self.certificate_path,
      'key': self.key_path,
      'certificate-expiration-days': '20'
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR Certificate '%s' will expire in less than 20 days" % (
        self.certificate_path,)
    )

  def test_expires_bad_value(self):
    self.createKeyCertificate(days=14)
    self.writePromise({
      'certificate': self.certificate_path,
      'key': self.key_path,
      'certificate-expiration-days': 'bad'
    })
    self.configureLauncher()
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertFailedMessage(
      self.getPromiseResult(self.promise_name),
      "ERROR certificate-expiration-days is wrong: 'bad'"
    )


class TestCheckCertificateSameFile(TestCheckCertificate):
  same_file = True
  pass


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
"""
Check if a process is running with an old version of the modules
found in the running python paths + optional additional python paths.

It parses all folders containing an `__init__.py` file and checks if
a file modification date is greater than the start date of the
process.
"""

from __future__ import print_function

import sys
import os
import errno
import argparse
import time
import psutil

ignored_extension_set = set([".pyc"])

def moduleIsModifiedSince(top, since, followlinks=False):
  for root, dir_list, file_list in os.walk(top, followlinks=followlinks):
    if root == top:
      continue
    if "__init__.py" not in file_list:
      del dir_list[:]
      continue
    for name in file_list:
      _, ext = os.path.splitext(name)
      if ext in ignored_extension_set:
        continue
      if since < os.stat(os.path.join(root, name)).st_mtime:
        print("%s was modified since the process started." %
                                                       os.path.join(root, name))
        print("Process Time %s < Last modified file %s" % (time.ctime(since),
                        time.ctime(os.stat(os.path.join(root, name)).st_mtime)))
        return True
  return False

def isProcessOlderThanDependencySet(pid, python_path_list, kill=False):
  process = psutil.Process(pid)
  start_time = process.create_time()
  if any(moduleIsModifiedSince(product_path, start_time) for product_path in python_path_list):
    if kill:
      print("Terminating process %s with pid %s" % (process.name(), pid))
      process.terminate()
    return True
  return False

def isProcessFromPidFileOlderThanDependencySet(pid_file_path, python_path_list, kill=False):
  with open(pid_file_path, "r") as f:
    pid = int(f.readline())
  return isProcessOlderThanDependencySet(pid, python_path_list, kill=kill)

def main():
  parser = argparse.ArgumentParser()
  parser.add_argument("-k", "--kill", action="store_true")
  parser.add_argument("pid_file_path", metavar="PID_FILE")
  parser.add_argument("python_path_list", nargs="*", metavar="ADDITIONAL_PYTHON_PATH", default=[])
  args = parser.parse_args()

  try:
    if isProcessFromPidFileOlderThanDependencySet(args.pid_file_path, sys.path + args.python_path_list, kill=args.kill):
      return 1
    return 0
  except (OSError, IOError) as err:
    if err.errno == errno.ENOENT:
      return 0
    raise
  except psutil.NoSuchProcess:
    return 0

if __name__ == "__main__":
  sys.exit(main())



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise
import time

import os
import sys
import re

r = re.compile(br"^([0-9]+\-[0-9]+\-[0-9]+ [0-9]+\:[0-9]+\:[0-9]+)(\,[0-9]+) - ([A-z]+) (.*)$")

@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(minute=10)

  def sense(self):
    log_file = self.getConfig('log-file')
    error_threshold = self.getConfig('error-threshold')
    maximum_delay = self.getConfig('maximum-delay')
    error_amount = 0
    if not os.path.exists(log_file):
      # file don't exist, nothing to check
      self.logger.info("log file does not exist: log check skipped")
      return 0
    
    with open(log_file, "rb") as f:
      f.seek(0, 2)
      block_end_byte = f.tell()
      f.seek(-min(block_end_byte, 4096*10), 1)
      data = f.read()
      for line in reversed(data.splitlines()):
        m = r.match(line)
        if m is None:
          continue
        dt, _, level, msg = m.groups()
        try:
          t = time.strptime(dt.decode('utf-8'), "%Y-%m-%d %H:%M:%S")
        except ValueError:
          continue
        if maximum_delay and (time.time()-time.mktime(t)) > maximum_delay:
          # no result in the latest hour
          break
        error_amount += 1
    if error_amount > error_threshold:
      self.logger.error('ERROR: Site has %s long request' % error_amount)
    else:
      self.logger.info('INFO: Site has %s long request' % error_amount)

  def test(self):
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    return self._test(result_count=3, failure_amount=3)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from .resiliencytestsuite import ResiliencyTestSuite

import logging
import random
import string
import time
from six.moves.urllib.request import urlopen

logger = logging.getLogger('KVMResiliencyTest')

def fetchKey(ip):
  """
  Fetch the key that had been set on original virtual hard drive.
  If doesn't exist (503), fail. If other error: retry after a few minutes,
  fail after XX (2?) hours.
  """
  new_key = None
  for i in range(0, 10):
    try:
      new_key = urlopen('http://%s:10080/get' % ip).read().strip()
      break
    except IOError:
      logger.error('Server in new KVM does not answer.')
      time.sleep(60)

  if not new_key:
    raise Exception('Server in new KVM does not answer for too long.')

  return new_key

class KVMTestSuite(ResiliencyTestSuite):
  """
  Run KVM Resiliency Test.
  Requires a specific KVM environment (virtual hard drive), see KVM SR for more
  informations.

  Scenario:
  1/ Boot from a custom image
  2/ The VM from the main instance starts a simple get/set server. It will receive a random number sent from the resiliency test. The VM will store this number, and the test suite will store the number as well.
  3/ Resilience is done, wait XX seconds
  4/ For each clone: do a takeover. Check that IPv6 of new main instance is different. Check, when doing a http request to the new VM that will fetch the stored random number, that the sent number is the same.

  Note: disk image is a simple debian with gunicorn and flask installed:
    apt-get install python-setuptools; easy_install gunicorn flask
  With the following python code running at boot in /root/number.py:

  import os

  from flask import Flask, abort, request
  app = Flask(__name__)

  storage = 'storage.txt'

  @app.route("/")
  def greeting_list(): # 'cause there are several greetings, and plural is forbidden.
    return "Hello World"

  @app.route("/get")
  def get():
    return open(storage, 'r').read()

  @app.route("/set")
  def set():
    #if os.path.exists(storage):
    #  abort(503)
    open(storage, 'w').write(request.args['key'])
    return "OK"

  if __name__ == "__main__":
    app.run(host='0.0.0.0', port=80)


  Then create the boot script:
  echo "cd /root; /usr/local/bin/gunicorn number:app -b 0.0.0.0:80 -D --error-logfile /root/error_log --access-logfile /root/access_log" > /etc/init.d/gunicorn-number
  chmod +x /etc/init.d/gunicorn-number
  update-rc.d gunicorn-number defaults


  There also is a script that randomly generates I/O in /root/io.sh:

  #!/bin/sh
  # Randomly generates high I/O on disk. Goal is to write on disk so that
  # it flushes at the same time that snapshot of disk image is done, to check if
  # it doesn't corrupt image.
  # Ayayo!
  while [ 1 ]; do
    dd if=/dev/urandom of=random count=2k
    sync
    sleep 0.2
  done

  Then create the boot script:
  echo "/bin/sh /root/io.sh &" > /etc/init.d/io
  chmod +x /etc/init.d/io
  update-rc.d io defaults

  """

  def _getPartitionParameterDict(self):
    """
    Overload default method.
    """
    return self.partition.request(
        software_release=self.software,
        software_type='kvm-resilient',
        partition_reference=self.root_instance_name).getConnectionParameterDict()

  def generateData(self):
    """
    Set a random key that will be stored inside of the virtual hard drive.
    """
    self.key = ''.join(random.SystemRandom().sample(string.ascii_lowercase, 20))
    self.logger.info('Generated key is: %s' % self.key)

  def pushDataOnMainInstance(self):
    self.logger.info('Getting the KVM IP...')
    self.ip = self._getPartitionParameterDict()['ipv6']
    logger.info('KVM IP is %s.' % self.ip)

    for i in range(0, 60):
      failure = False
      try:
        connection = urlopen('http://%s:10080/set?key=%s' % (self.ip, self.key))
        if connection.getcode() is 200:
          break
        else:
          failure = True
      except IOError:
        failure = True
      finally:
        if failure:
          logger.info('Impossible to connect to virtual machine to set key. sleeping...')
          time.sleep(60)
      if i is 59:
        raise Exception('Bad return code when setting key in main instance, after trying for 60 minutes.')

    logger.info('Key uploaded to KVM main instance.')

  def checkDataOnCloneInstance(self):
    self.ip = self._returnNewInstanceParameter(
        parameter_key='ipv6',
        old_parameter_value=self.ip
    )

    new_key = fetchKey(self.ip)
    logger.info('Key on this new instance is %s' % new_key)

    # Compare with original key. If same: success.
    if new_key == self.key:
      self.logger.info('Data are the same: success.')
      return True
    else:
      self.logger.info('Data are different: failure.')


def runTestSuite(*args, **kwargs):
  """
  Run KVM Resiliency Test.
  """
  return KVMTestSuite(*args, **kwargs).runTestSuite()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import subprocess
import json
import psutil
import time
from datetime import datetime
from shutil import copyfile
import glob
import argparse
import traceback
import logging
from six.moves import configparser
from slapos.grid.promise import PromiseLauncher, PromiseQueueResult, PromiseError
from slapos.grid.promise.generic import PROMISE_LOG_FOLDER_NAME
from slapos.util import mkdir_p

# Promise timeout after 20 seconds by default
promise_timeout = 20

def getArgumentParser():
  """
  Parse arguments for monitor collector instance.
  """
  parser = argparse.ArgumentParser()
  parser.add_argument('-c', '--config', dest='config_file',
                      help='The Path of configuration file to load.')
  parser.add_argument('-p', '--partition-folder',
                      help='Base path of the partition.')
  parser.add_argument('-L', '--log-folder',
                      help='Folder where promises will write logs.')
  parser.add_argument('--console',
                      help='Log to console too',
                      default=False, action='store_true')
  parser.add_argument('-a', '--check-anomaly',
                      help='Enable check of anomaly on promises.',
                      default=False, action='store_true')
  parser.add_argument('-D', '--debug',
                      help='Configure loggin in debug mode.',
                      default=False, action='store_true')
  parser.add_argument('--master-url',
                      help='SlapOS Master Service URL.')
  parser.add_argument('--partition-cert',
                      help='Computer partition certificate file path.')
  parser.add_argument('--partition-key',
                      help='Computer partition key file path.')
  parser.add_argument('--partition-id',
                      help='Computer partition ID.')
  parser.add_argument('--computer-id',
                      help='Computer ID.')
  parser.add_argument('--pid-path',
                      help='Path where the pid of this process will be writen.')
  parser.add_argument('--run-only',
                      help='List of python promises to run separates by space.',
                      default="")
  parser.add_argument('-f', '--force',
                      help='Force to run promise without consider the periodicity.',
                      default=False, action='store_true')
  parser.add_argument('--dry-run',
                      help='Only run promise without save result.',
                      default=False, action='store_true')
  parser.add_argument('--promise-timeout', default=20, type=int,
                      help='Maximum promise execution time.')

  return parser


class MonitorPromiseLauncher(object):

  def __init__(self, config_parser):
    self.config = config_parser
    if self.config.config_file:
      self._loadConfigFromFile(self.config.config_file)
    self.logger = logging.getLogger("Monitor")
    self.logger.setLevel(logging.DEBUG if self.config.debug else logging.INFO)

    if not self.config.log_folder or self.config.console:
      if len(self.logger.handlers) == 0 or \
          not isinstance(self.logger.handlers[0], logging.StreamHandler):
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
        self.logger.addHandler(handler)
    if self.config.log_folder:
      log_file = os.path.join(self.config.log_folder,
        '%s.log' % self.config.title.replace(' ', '_'))
      file_handler = logging.FileHandler(log_file)
      file_handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
      self.logger.addHandler(file_handler)


  def _loadConfigFromFile(self, config_file):
    config = configparser.ConfigParser()
    config.read([config_file])
    known_key_list = ['partition-cert', 'partition-key', 'partition-id',
                      'pid-path', 'computer-id', 'check-anomaly',
                      'master-url', 'partition-folder', 'promise-timeout']
 
    if config.has_section('promises'):
      for key, value in config.items('promises'):
        key_add = key.replace('-', '_')
        if key in known_key_list and \
            getattr(self.config, key_add, None) is None:
          setattr(self.config, key_add, value)

  def start(self):
    """
      run all promises in sequential ways
    """
    if self.config.pid_path:
      if os.path.exists(self.config.pid_path):
        # Check if another run promise is running
        with open(self.config.pid_path) as fpid:
          try:
            pid = int(fpid.read(6))
          except ValueError:
            pid = None
          if pid and os.path.exists("/proc/" + str(pid)):
            self.logger.warning("A process is already running with pid " + str(pid))
            return []

      with open(self.config.pid_path, 'w') as fpid:
        fpid.write(str(os.getpid()))

    if not self.config.partition_folder:
      raise ValueError("Partition folder is not specified")

    parameter_dict = {
      'promise-timeout': self.config.promise_timeout or promise_timeout,
      'promise-folder': os.path.join(self.config.partition_folder, 'etc', 'plugin'),
      'legacy-promise-folder': os.path.join(self.config.partition_folder, 'etc', 'promise'),
      'partition-folder': self.config.partition_folder,
      'master-url': self.config.master_url,
      'partition-cert': self.config.partition_cert,
      'partition-key': self.config.partition_key,
      'partition-id': self.config.partition_id,
      'computer-id': self.config.computer_id,
      'debug': self.config.debug,
      'check-anomaly': self.config.check_anomaly,
      'force': self.config.force,
      'run-only-promise-list': [x for x in self.config.run_only.split(' ') if x]
    }
    if self.config.log_folder:
      parameter_dict['log-folder'] = self.config.log_folder
    else:
      parameter_dict['log-folder'] = os.path.join(self.config.partition_folder,
                                                  PROMISE_LOG_FOLDER_NAME)
      mkdir_p(parameter_dict['log-folder'])

    promise_launcher = PromiseLauncher(
      config=parameter_dict,
      logger=self.logger,
      dry_run=self.config.dry_run
    )

    self.logger.info("Checking promises...")
    exit_code = 0
    try:
      promise_launcher.run()
    except PromiseError as e:
      # error was already logged
      exit_code = 1
    if self.config.pid_path:
      os.remove(self.config.pid_path)
    self.logger.info("Finished promises.")
    return exit_code

def main():
  arg_parser = getArgumentParser()
  promise_runner = MonitorPromiseLauncher(arg_parser.parse_args())
  sys.exit(promise_runner.start())



# -*- coding: utf-8 -*-
# vim: set et sts=2:
# pylint: disable-msg=W0311,C0301,C0103,C0111,R0904

#############################################
# !!! Attention !!!
# You now have to comment the last line
# in __init__py, wich starts the functiun
# run() in order to start the tests,
# or it will NOT work
#############################################

from __future__ import print_function

import argparse
import base64
from six.moves.configparser import SafeConfigParser
import datetime
import hashlib
import json
import os
import shutil
from . import sup_process
from io import StringIO
import ssl
import time
import unittest
from six.moves.urllib.request import Request, urlopen
import six

from slapos.runner.utils import (getProfilePath,
                                 getSession, isInstanceRunning,
                                 isSoftwareRunning, startProxy,
                                 isSoftwareReleaseReady,
                                 runSlapgridUntilSuccess, runInstanceWithLock,
                                 getBuildAndRunParams, saveBuildAndRunParams)
from slapos.runner import views
import slapos.slap
from slapos.htpasswd import  HtpasswdFile
from .run import Config

import erp5.util.taskdistribution as taskdistribution


#Helpers
def loadJson(response):
  return json.loads(response.data)

def parseArguments():
  """
  Parse arguments for erp5testnode-backed test.
  """
  parser = argparse.ArgumentParser()

  # runnertest mandatory arguments
  parser.add_argument('--key_file',
                      metavar='KEY_FILE',
                      help='Path to the key file used to communicate with slapOS-master')

  parser.add_argument('--cert_file',
                      metavar='CERT_FILE',
                      help='Path to the cert file used to communicate with slapOS-master')

  parser.add_argument('--server_url',
                      metavar='SERVER_URL',
                      help='URL of the local slapproxy')

  parser.add_argument('--computer_id',
                      metavar='COMPUTER_ID',
                      help='ID of the COMP where the slaprunner is running')

  parser.add_argument('--partition_id',
                      metavar='PARTITION_ID',
                      help='ID of the partition where the slaprunner is depoyed')

  # Test Node arguments
  parser.add_argument('--test_result_path',
                      metavar='ERP5_TEST_RESULT_PATH',
                      help='ERP5 relative path of the test result')

  parser.add_argument('--revision',
                      metavar='REVISION',
                      help='Revision of the test_suite')

  parser.add_argument('--test_suite',
                      metavar='TEST_SUITE',
                      help='Name of the test suite')

  parser.add_argument('--test_suite_title',
                      metavar='TEST_SUITE',
                      help='The test suite title')

  parser.add_argument('--test_node_title',
                      metavar='NODE_TITLE',
                      help='Title of the testnode which is running this'
                            'launcher')

  parser.add_argument('--project_title',
                      metavar='PROJECT_TITLE',
                      help='The project title')

  parser.add_argument('--node_quantity', help='Number of parallel tests to run',
                      default=1, type=int)

  parser.add_argument('--master_url',
                      metavar='TEST_SUITE_MASTER_URL',
                      help='Url to connect to the ERP5 Master testsuite taskditributor')

  return parser.parse_args()


class SlaprunnerTestCase(unittest.TestCase):

  @classmethod
  def setUpClass(cls, **kw):
    if len(kw) == 0:
      return
    cls.server_url = kw['server_url']
    cls.key_file = kw['key_file']
    cls.cert_file = kw['cert_file']
    cls.computer_id = kw['computer_id']
    cls.partition_id = kw['partition_id']
    # Get parameters returned by slapos master
    slap = slapos.slap.slap()
    slap.initializeConnection(cls.server_url, cls.key_file, cls.cert_file)
    cls.partition = slap.registerComputerPartition(
      computer_guid=cls.computer_id,
      partition_id=cls.partition_id
    )
    cls.parameter_dict = cls.partition.getConnectionParameterDict()
    for attribute, value in six.iteritems(cls.parameter_dict):
      setattr(cls, attribute.replace('-', '_'), value)

    #create slaprunner configuration
    views.app.config['TESTING'] = True
    config = Config()
    config.setConfig()
    views.app.config.update(**config.__dict__)
    cls.app = views.app.test_client()
    cls.app.config = views.app.config

    # Set up path (needed to find git binary)
    os.environ['PATH'] = config.path

  def setUp(self):
    """Initialize slapos webrunner here"""
    self.users = [self.init_user, self.init_password, "slaprunner@nexedi.com", "SlapOS web runner"]
    self.updateUser = ["newslapuser", "newslappwd", "slaprunner@nexedi.com", "SlapOS web runner"]
    self.repo = 'https://lab.nexedi.com/rafael/slapos-workarround.git'
    self.software = "workspace/slapos/software/"  # relative directory fo SR
    self.project = 'slapos'  # Default project name
    self.template = 'template.cfg'
    self.partitionPrefix = 'slappart'

    self.workdir = workdir = os.path.join(self.app.config['runner_workdir'], 'project')
    software_link = os.path.join(self.app.config['runner_workdir'], 'softwareLink')
    #update or create all runner base directory to test_dir
    if not os.path.exists(workdir):
      os.mkdir(workdir)
    if not os.path.exists(software_link):
      os.mkdir(software_link)

    #Create config.json
    json_file = os.path.join(views.app.config['etc_dir'], 'config.json')
    if not os.path.exists(json_file):
      params = {
        'run_instance' : True,
        'run_software' : True,
        'max_run_instance' : 3,
        'max_run_software' : 2
      }
      open(json_file, "w").write(json.dumps(params))

  def tearDown(self):
    """Remove all test data"""
    project = os.path.join(self.app.config['etc_dir'], '.project')

    #reset tested parameters
    self.updateConfigParameter('autorun', False)
    self.updateConfigParameter('auto_deploy', True)

    if os.path.exists(project):
      os.unlink(project)
    if os.path.exists(self.app.config['workspace']):
      shutil.rmtree(self.app.config['workspace'])
    if os.path.exists(self.app.config['software_root']):
      shutil.rmtree(self.app.config['software_root'])
    if os.path.exists(self.app.config['instance_root']):
      shutil.rmtree(self.app.config['instance_root'])
    if os.path.exists(self.app.config['software_link']):
      shutil.rmtree(self.app.config['software_link'])

  def updateConfigParameter(self, parameter, value):
    config_parser = SafeConfigParser()
    config_parser.read(os.getenv('RUNNER_CONFIG'))
    for section in config_parser.sections():
      if config_parser.has_option(section, parameter):
        config_parser.set(section, parameter, str(value))
    with open(os.getenv('RUNNER_CONFIG'), 'wb') as configfile:
      config_parser.write(configfile)


  def updateAccount(self, newaccount):
    """Helper for update user account data"""
    return self.app.post('/updateAccount',
                         data=dict(
                           username=newaccount[0],
                           password=newaccount[1],
                           email=newaccount[2],
                           name=newaccount[3],
                         ),
                         follow_redirects=True)

  def getCurrentSR(self):
   return getProfilePath(self.app.config['etc_dir'],
                              self.app.config['software_profile'])

  def setupProjectFolder(self):
    """Helper to create a project folder as for slapos.git"""
    base = os.path.join(self.app.config['workspace'], 'slapos')
    software = os.path.join(base, 'software')
    os.mkdir(base)
    os.mkdir(software)

  def setupTestSoftware(self):
    """Helper to setup Basic SR for testing purposes"""
    self.setupProjectFolder()
    base = os.path.join(self.app.config['workspace'], 'slapos')
    software = os.path.join(base, 'software')
    testSoftware = os.path.join(software, 'slaprunner-test')
    sr = "[buildout]\n\n"
    sr += "parts = command\n\nunzip = true\nnetworkcache-section = networkcache\n\n"
    sr += "find-links += http://www.nexedi.org/static/packages/source/slapos.buildout/\n\n"
    sr += "[networkcache]\ndownload-cache-url = http://www.shacache.org/shacache"
    sr += "\ndownload-dir-url = http://www.shacache.org/shadir\n\n"
    sr += "[command]\nrecipe = zc.recipe.egg\neggs = plone.recipe.command\n  zc.buildout\n\n"
    sr += """
[versions]
setuptools = 33.1.1
"""
    os.mkdir(testSoftware)
    open(os.path.join(testSoftware, self.app.config['software_profile']),
         'w').write(sr)
    md5 = hashlib.md5(os.path.join(self.app.config['workspace'],
                                   "slapos/software/slaprunner-test",
                                   self.app.config['software_profile'])
                      ).hexdigest()
    base = os.path.join(self.app.config['software_root'], md5)
    template = os.path.join(base, self.template)
    content = "[buildout]\n"
    content += "parts = \n  create-file\n\n"
    content += "eggs-directory = %s\n" % os.path.join(base, 'eggs')
    content += "develop-eggs-directory = %s\n\n" % os.path.join(base, 'develop-eggs')
    content += "[create-file]\nrecipe = plone.recipe.command\n"
    content += "filename = ${buildout:directory}/etc\n"
    content += "command = mkdir ${:filename} && echo 'simple file' > ${:filename}/testfile\n"
    os.mkdir(self.app.config['software_root'])
    os.mkdir(base)
    open(template, "w").write(content)

  def assertCanLoginWith(self, username, password):
    request = Request(self.backend_url)
    base64string = base64.encodestring('%s:%s' % (username, password))[:-1]
    request.add_header("Authorization", "Basic %s" % base64string)
    ssl_context = ssl._create_unverified_context()
    result = urlopen(request, context=ssl_context)
    self.assertEqual(result.getcode(), 200)

  def test_updateAccount(self):
    """test Update accound, this needs the user to log in"""
    # Check that given user can log in
    self.assertCanLoginWith(self.users[0], self.users[1])
    # Adds a user
    new_user = {'username': 'vice-president', 'password': '123456'}
    response = loadJson(self.app.post('/addUser', data=new_user))
    self.assertEqual(response['code'], 1)
    self.assertCanLoginWith(new_user['username'], new_user['password'])


  def test_cloneProject(self):
    """Start scenario 1 for deploying SR: Clone a project from git repository"""
    folder = 'workspace/' + self.project
    if os.path.exists(self.app.config['workspace'] + '/' + self.project):
      shutil.rmtree(self.app.config['workspace'] + '/' + self.project)
    data = {
      'repo': self.repo,
      'user': 'Slaprunner test',
      'email': 'slaprunner@nexedi.com',
      'name': folder
    }
    response = loadJson(self.app.post('/cloneRepository', data=data,
                    follow_redirects=True))
    self.assertEqual(response['result'], "")
    # Get realpath of create project
    path_data = dict(file=folder)
    response = loadJson(self.app.post('/getPath', data=path_data,
                    follow_redirects=True))
    self.assertEqual(response['code'], 1)
    realFolder = response['result'].split('#')[0]
    #Check git configuration
    config = open(os.path.join(realFolder, '.git/config')).read()
    assert "slaprunner@nexedi.com" in config and "Slaprunner test" in config

    # Checkout to slaprunner branch, this supposes that branch slaprunner exit
    response = loadJson(self.app.post('/newBranch',
                                      data=dict(
                                        project=folder,
                                        create='0',
                                        name='erp5'
                                      ),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")

  def test_createSR(self):
    """Scenario 2: Create a new software release"""
    #setup project directory
    self.setupProjectFolder()
    newSoftware = os.path.join(self.software, 'slaprunner-test')
    response = loadJson(self.app.post('/createSoftware',
                                      data=dict(folder=newSoftware),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")
    currentSR = self.getCurrentSR()
    assert newSoftware in currentSR

  def test_openSR(self):
    """Scenario 3: Open software release"""
    self.test_cloneProject()
    software = os.path.join(self.software, 'helloworld')  # Drupal SR must exist in SR folder
    response = loadJson(self.app.post('/setCurrentProject',
                                      data=dict(path=software),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")
    currentSR = self.getCurrentSR()
    assert software in currentSR
    self.assertFalse(isInstanceRunning(self.app.config))
    self.assertFalse(isSoftwareRunning(self.app.config))


  def test_runSoftware(self):
    """Scenario 4: CReate empty SR and save software.cfg file
      then run slapgrid-sr
    """
    #Call config account
    #call create software Release
    self.test_createSR()
    newSoftware = self.getCurrentSR()
    softwareRelease = "[buildout]\n\nparts =\n  test-application\n"
    softwareRelease += "#Test download git web repos éè@: utf-8 caracters\n"
    softwareRelease += "[test-application]\nrecipe = hexagonit.recipe.download\n"
    softwareRelease += "url = https://lab.nexedi.com/rafael/slapos-workarround.git\n" 
    softwareRelease += "filename = slapos.git\n"
    softwareRelease += "download-only = true\n"
    softwareRelease += "[versions]\nsetuptools = 33.1.1"
    response = loadJson(self.app.post('/saveFileContent',
                                      data=dict(file=newSoftware,
                                                content=softwareRelease),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")

    # Compile software and wait until slapgrid ends
    # this is supposed to use current SR
    while self.app.get('/isSRReady').data == "2":
      time.sleep(2)
    self.assertEqual(self.app.get('/isSRReady').data, "1")
    self.assertTrue(os.path.exists(self.app.config['software_root']))
    self.assertTrue(os.path.exists(self.app.config['software_log']))
    assert "test-application" in open(self.app.config['software_log']).read()
    sr_dir = os.listdir(self.app.config['software_root'])
    self.assertEqual(len(sr_dir), 1)
    createdFile = os.path.join(self.app.config['software_root'], sr_dir[0],
                              'parts', 'test-application', 'slapos.git')
    self.assertTrue(os.path.exists(createdFile))


  def test_updateInstanceParameter(self):
    """Scenario 5: Update parameters of current sofware profile"""
    self.setupTestSoftware()
    #Set current projet and run Slapgrid-cp
    software = self.software + 'slaprunner-test/'
    response = loadJson(self.app.post('/setCurrentProject',
                                      data=dict(path=software),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")
    #Send paramters for the instance
    parameterDict = dict(appname='slaprunnerTest', cacountry='France')
    parameterXml = '<?xml version="1.0" encoding="utf-8"?>\n<instance>'
    parameterXml += '<parameter id="appname">slaprunnerTest</parameter>\n'
    parameterXml += '<parameter id="cacountry">France</parameter>\n</instance>'
    software_type = 'production'
    response = loadJson(self.app.post('/saveParameterXml',
                                      data=dict(parameter=parameterXml,
                                                software_type=software_type),
                                      follow_redirects=True))
    self.assertEqual(response['result'], "")
    slap = slapos.slap.slap()
    slap.initializeConnection(self.app.config['master_url'])
    computer = slap.registerComputer(self.app.config['computer_id'])
    partitionList = computer.getComputerPartitionList()
    self.assertNotEqual(partitionList, [])
    #Assume that the requested partition is partition 0
    slapParameterDict = partitionList[0].getInstanceParameterDict()
    self.assertTrue('appname' in slapParameterDict)
    self.assertTrue('cacountry' in slapParameterDict)
    self.assertEqual(slapParameterDict['appname'], 'slaprunnerTest')
    self.assertEqual(slapParameterDict['cacountry'], 'France')
    self.assertEqual(slapParameterDict['slap_software_type'], 'production')

    #test getParameterXml for webrunner UI
    response = loadJson(self.app.get('/getParameterXml/xml'))
    self.assertEqual(parameterXml, response['result'])
    response = loadJson(self.app.get('/getParameterXml/dict'))
    self.assertEqual(parameterDict, response['result']['instance'])


  def test_requestInstance(self):
    """Scenario 6: request software instance"""
    self.test_updateInstanceParameter()
    #run Software profile
    response = loadJson(self.app.post('/runSoftwareProfile',
                                      data=dict(),
                                      follow_redirects=True))
    while self.app.get('/isSRReady').data == "2":
      time.sleep(2)
    self.assertEqual(self.app.get('/isSRReady').data, "1")
    #run instance profile
    response = loadJson(self.app.post('/runInstanceProfile',
                                      data=dict(),
                                      follow_redirects=True))
    self.assertTrue(response['result'])
    # lets some time to the Instance to be deployed
    time.sleep(5)
    #Check that all partitions has been created
    assert "create-file" in open(self.app.config['instance_log']).read()
    for num in range(int(self.app.config['partition_amount'])):
      partition = os.path.join(self.app.config['instance_root'],
                    self.partitionPrefix + str(num))
      self.assertTrue(os.path.exists(partition))

    #Go to partition 0
    instancePath = os.path.join(self.app.config['instance_root'],
                         self.partitionPrefix + '0')
    createdFile = os.path.join(instancePath, 'etc', 'testfile')
    self.assertTrue(os.path.exists(createdFile))
    assert 'simple file' in open(createdFile).read()


  def test_safeAutoDeploy(self):
    """Scenario 7: isSRReady won't overwrite the existing
    Sofware Instance if it has been deployed yet"""
    # Test that SR won't be deployed with auto_deploy=False
    self.updateConfigParameter('auto_deploy', False)
    self.updateConfigParameter('autorun', False)
    project = open(os.path.join(self.app.config['etc_dir'],
                  '.project'), "w")
    project.write(self.software + 'slaprunner-test/')
    project.close()
    response = isSoftwareReleaseReady(self.app.config)
    self.assertEqual(response, "0")
    # Test if auto_deploy parameter starts the deployment of SR
    self.updateConfigParameter('auto_deploy', True)
    self.setupTestSoftware()
    response = isSoftwareReleaseReady(self.app.config)
    self.assertEqual(response, "2")
    # Test that the new call to isSoftwareReleaseReady
    # doesn't overwrite the previous installed one
    sup_process.killRunningProcess(self.app.config, 'slapgrid-sr')
    completed_path = os.path.join(self.app.config['runner_workdir'],
        'softwareLink', 'slaprunner-test', '.completed')
    completed_text = ".completed file: test"
    completed = open(completed_path, "w")
    completed.write(completed_text)
    completed.close()
    response = isSoftwareReleaseReady(self.app.config)
    self.assertEqual(response, "1")
    assert completed_text in open(completed_path).read()

  def test_maximumRunOfSlapgrid(self):
    """Scenario 8: runSlapgridUntilSucces run until a defined maximum of time
    slapgrid-sr and slapgrid-cp if it fails. It can also run only one or both
    of them if it is defined so
    We directly calls runSlapgridUntilSuccess, because we want
    to test the return code of the function"""
    # Installs a wrong buildout which will fail
    MAX_RUN_SOFTWARE = getBuildAndRunParams(self.app.config)['max_run_software']
    MAX_RUN_INSTANCE = getBuildAndRunParams(self.app.config)['max_run_instance']
    self.test_createSR()
    newSoftware = self.getCurrentSR()
    softwareRelease = "[buildout]\n\nparts =\n  test-application\n"
    softwareRelease += "find-links += http://www.nexedi.org/static/packages/source/slapos.buildout/\n\n"
    softwareRelease += "[networkcache]\ndownload-cache-url = http://www.shacache.org/shacache"
    softwareRelease += "\ndownload-dir-url = http://www.shacache.org/shadir\n\n"
    softwareRelease += "#Test download git web repos éè@: utf-8 caracters\n"
    softwareRelease += "[test-application]\nrecipe = slapos.cookbook:mkdirectory\n"
    softwareRelease += "test = /root/test\n"
    softwareRelease += """
[versions]
setuptools = 33.1.1
"""
    response = loadJson(self.app.post('/saveFileContent',
                                      data=dict(file=newSoftware,
                                                content=softwareRelease),
                                      follow_redirects=True))
    response = runSlapgridUntilSuccess(self.app.config, 'software')
    self.assertEqual(response, MAX_RUN_SOFTWARE)
    # clean folders for other tests
    workdir = os.path.join(self.app.config['runner_workdir'], 'project')
    git_repo = os.path.join(workdir, 'slapos')
    if os.path.exists(git_repo):
      shutil.rmtree(git_repo)
    # Installs a software which deploys, but fails while instanciating
    # preparation
    base = os.path.join(self.app.config['workspace'], 'slapos')
    software = os.path.join(base, 'software')
    testSoftware = os.path.join(software, 'slaprunner-test')
    if not os.path.exists(testSoftware):
      os.makedirs(testSoftware)
    software_cfg = os.path.join(testSoftware, 'software.cfg')
    instance_cfg = os.path.join(testSoftware, 'instance.cfg')
    # software.cfg
    softwareRelease = "[buildout]\n\nparts =\n  failing-template\n\n"
    softwareRelease += "[failing-template]\nrecipe = hexagonit.recipe.download\n"
    softwareRelease += "url = %s\n" % (instance_cfg)
    softwareRelease += "destination = ${buildout:directory}\n"
    softwareRelease += "download-only = true\n"
    open(software_cfg, 'w+').write(softwareRelease)
    # instance.cfg
    content = "[buildout]\n\nparts =\n fail\n"
    content += "[fail]\nrecipe=plone.recipe.command\n"
    content += "command = exit 1"
    open(instance_cfg, 'w+').write(content)
    project = open(os.path.join(self.app.config['etc_dir'],
                  '.project'), "w")
    project.write(self.software + 'slaprunner-test')
    project.close()
    # Build and Run
    parameters = getBuildAndRunParams(self.app.config)
    parameters['run_instance'] = False
    saveBuildAndRunParams(self.app.config, parameters)
    response = runSlapgridUntilSuccess(self.app.config, 'software')
    self.assertEqual(response, 1)
    parameters['run_instance'] = True
    saveBuildAndRunParams(self.app.config, parameters)
    response = runSlapgridUntilSuccess(self.app.config, 'software')
    self.assertEqual(response, (1, MAX_RUN_INSTANCE))

  def test_slaveInstanceDeployment(self):
    """
    In order to test both slapproxy and core features of
    slaprunner, will install special Software Release
    into the current webrunner and fetch its instance
    parameters once deployed.
    """
    # XXX: This test should NOT be a unit test but should be run
    # by a Test Agent running against a slapproxy.

    # Deploy "test-slave-instance-deployment" Software Release
    self.test_cloneProject()
    software = os.path.join(self.software, 'test-slave-instance-deployment')

    # Checkout to master branch
    response = loadJson(self.app.post('/newBranch',
                                      data=dict(
                                        project=self.workdir+'/slapos',
                                        create='0',
                                        name='master'
                                      ),
                                      follow_redirects=True)).get(u'code')
    self.assertEqual(response, 1)
    response = loadJson(self.app.post('/setCurrentProject',
                                      data=dict(path=software),
                                      follow_redirects=True)).get(u'code')
    self.assertEqual(response, 1)

    response = loadJson(self.app.post('/saveParameterXml',
                                      data=dict(parameter='<?xml version="1.0" encoding="utf-8"?>\n<instance/>',
                                                software_type='default'),
                                      follow_redirects=True))

    while self.app.get('/isSRReady').data == "2":
      time.sleep(2)
    self.assertEqual(self.app.get('/isSRReady').data, "1")

    # Run instance deployment 3 times
    runInstanceWithLock(self.app.config)
    runInstanceWithLock(self.app.config)
    result = runInstanceWithLock(self.app.config)
    # result is True if returncode is 0 (i.e "deployed and promise passed")
    self.assertTrue(result)


  def test_dynamicParametersReading(self):
    """Test if the value of a parameter can change in the flask application
    only by changing the value of slapos.cfg config file. This can happen when
    slapgrid processes the webrunner's partition.
    """
    config_file = os.path.join(self.app.config['etc_dir'], 'slapos-test.cfg')
    runner_config_old = os.environ['RUNNER_CONFIG']
    os.environ['RUNNER_CONFIG'] = config_file
    open(config_file, 'w').write("[section]\nvar=value")
    config = self.app.config
    self.assertEqual(config['var'], "value")
    open(config_file, 'w').write("[section]\nvar=value_changed")
    self.assertEqual(config['var'], "value_changed")
    # cleanup
    os.environ['RUNNER_CONFIG'] = runner_config_old


class PrintStringIO(StringIO):
  def write(self, data):
    StringIO.write(self, data)
    print(data)

def main():
  """
  Function meant to be run by erp5testnode.
  """
  args = parseArguments()
  master = taskdistribution.TaskDistributor(args.master_url)
  test_suite_title = args.test_suite_title or args.test_suite
  revision = args.revision

  SlaprunnerTestCase.setUpClass(
    key_file=args.key_file,
    cert_file=args.cert_file,
    server_url=args.server_url,
    computer_id=args.computer_id,
    partition_id=args.partition_id)

  test_result = master.createTestResult(revision, [test_suite_title],
    args.test_node_title, True, test_suite_title, args.project_title)

  if test_result is None:
    # Thereis nothing to run here, all tests are been running by
    # some other node.
    return
  test_line = test_result.start()

  start_time = time.time()

  stderr = PrintStringIO()
  suite = unittest.TestLoader().loadTestsFromTestCase(SlaprunnerTestCase)
  test_result = unittest.TextTestRunner(verbosity=2, stream=stderr).run(suite)

  test_duration = time.time() - start_time
  test_line.stop(stderr=stderr.getvalue(),
                  test_count=test_result.testsRun,
                  error_count=len(test_result.errors),
                  failure_count=len(test_result.failures) + len(test_result.unexpectedSuccesses),
                  skip_count=len(test_result.skipped),
                  duration=test_duration)

def runStandaloneUnitTest():
  """
  Run unit tests without erp5testnode."
  """
  unittest.main(module=__name__)



import csv
import feedparser
import os
import shutil
import tempfile
import unittest

from slapos.resilient.rdiffBackupStat2RSS import genRSS

class Option(dict):
    def __init__(self, **kw): 
        self.__dict__.update(kw)
    def __setitem__(i, y):
        self.__dict__[i] = y

class TestRdiffBackupStat2RSS(unittest.TestCase):
  def setUp(self):
    self.data_directory = tempfile.mkdtemp(dir='.')
    self.feed_path = os.path.join(self.data_directory)

  def tearDown(self):
    shutil.rmtree(self.data_directory)

  def getOptionObject(self, **kw):
    """
    Returns an object containing options as properties, to simulate a call
    to the tested script
    """
    option = {
      'rdiff_backup_data_folder': self.data_directory,
      'output': self.feed_path,
      'feed_url': 'http://exemple.com',
    }
    option.update(kw)
    return Option(**option)

  def createSample(self):
    """
    Writes 2 statistics file in rdiff-backup format
    """
    with open(os.path.join(self.data_directory, 'session_statistics_1'), 'w') as stat_file:
      stat_file.write("""\
        StartTime 1473339659.00 (Thu Sep  8 15:00:59 2016)
        EndTime 1473339667.81 (Thu Sep  8 15:01:07 2016)
        ElapsedTime 8.81 (8.81 seconds)
        SourceFiles 2381
        SourceFileSize 142096473 (136 MB)
        MirrorFiles 1
        MirrorFileSize 0 (0 bytes)
        NewFiles 2380
        NewFileSize 142096473 (136 MB)
        DeletedFiles 0
        DeletedFileSize 0 (0 bytes)
        ChangedFiles 1
        ChangedSourceSize 0 (0 bytes)
        ChangedMirrorSize 0 (0 bytes)
        IncrementFiles 0
        IncrementFileSize 0 (0 bytes)
        TotalDestinationSizeChange 142096473 (136 MB)
        Errors 0""")
    with open(os.path.join(self.data_directory, 'session_statistics_2'), 'w') as stat_file:
      stat_file.write("""\
        StartTime 1473340154.00 (Thu Sep  8 15:09:14 2016)
        EndTime 1473340154.95 (Thu Sep  8 15:09:14 2016)
        ElapsedTime 0.95 (0.95 seconds)
        SourceFiles 2381
        SourceFileSize 142096473 (136 MB)
        MirrorFiles 2381
        MirrorFileSize 142096473 (136 MB)
        NewFiles 0
        NewFileSize 0 (0 bytes)
        DeletedFiles 0
        DeletedFileSize 0 (0 bytes)
        ChangedFiles 15
        ChangedSourceSize 230112 (225 KB)
        ChangedMirrorSize 230112 (225 KB)
        IncrementFiles 15
        IncrementFileSize 2122 (2.07 KB)
        TotalDestinationSizeChange 2122 (2.07 KB)
        Errors 0""")

  def test_generatedRSSIsCorrect(self):
    self.createSample()
    option = self.getOptionObject()
    feed_content = genRSS(option)
    feed = feedparser.parse(feed_content)

    self.assertFalse(feed.bozo)
    self.assertTrue(len(feed.entries), 2)
    self.assertLess(feed.entries[0].published_parsed, feed.entries[1].published_parsed)


if __name__ == '__main__':
  unittest.main()



"""
Some notable parameters:

  promise-timeout:
    Optional timeout (in seconds) for promise.
  timeout:
    Optional timeout (in seconds) for HTTP request.
  verify, ca-cert-file, cert-file, key-file:
    Optional SSL information. (See Python requests documentation.)
  http-code:
    (default 200) The expected response HTTP code.
  ignore-code:
    (default 0) If set to 1, ignore the response HTTP code.
  username, password:
    If supplied, enables basic HTTP authentication.
"""

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

import requests


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # SR can set custom periodicity
    self.setPeriodicity(float(self.getConfig('frequency', 2)))

  def sense(self):
    """
      Check if frontend URL is available.
    """

    url = self.getConfig('url')
    # make default time a max of 5 seconds, a bit smaller than promise-timeout
    # and in the same time at least 1 second
    default_timeout = max(
      1, min(5, int(self.getConfig('promise-timeout', 20)) - 1))
    expected_http_code = int(self.getConfig('http-code', 200))
    ca_cert_file = self.getConfig('ca-cert-file')
    cert_file = self.getConfig('cert-file')
    key_file = self.getConfig('key-file')
    verify = int(self.getConfig('verify', 0))
    username = self.getConfig('username')
    password = self.getConfig('password')

    if int(self.getConfig('ignore-code', 0)) == 1:
      ignore_code = True
    else:
      ignore_code = False

    if ca_cert_file:
      verify = ca_cert_file
    elif verify:
      verify = True
    else:
      verify = False

    if key_file and cert_file:
      cert = (cert_file, key_file)
    else:
      cert = None

    if username and password:
      credentials = (username, password)
      request_type = "authenticated"
    else:
      credentials = None
      request_type = "non-authenticated"

    request_options = {
      'allow_redirects': True,
      'timeout': int(self.getConfig('timeout', default_timeout)),
      'verify': verify,
      'cert': cert,
      'auth': credentials,
    }

    try:
      response = requests.get(url, **request_options)
    except requests.exceptions.SSLError as e:
      if 'certificate verify failed' in str(e):
        self.logger.error(
          "ERROR SSL verify failed while accessing %r", url)
      else:
        self.logger.error(
          "ERROR Unknown SSL error %r while accessing %r", e, url)
    except requests.ConnectionError as e:
      self.logger.error(
        "ERROR connection not possible while accessing %r", url)
    except Exception as e:
      self.logger.error("ERROR: %s", e)

    else:
      # Log a sensible message, depending on the request/response
      # parameters.
      if ignore_code:
        log = self.logger.info
        result = "succeeded"
        message = "return code ignored"
      elif response.status_code == expected_http_code:
        log = self.logger.info
        result = "succeeded"
        message = "returned expected code %d" % expected_http_code
      else:
        log = self.logger.error
        result = "failed"
        message = "returned %d, expected %d" % (response.status_code,
                                                expected_http_code)

      log("%s request to %r %s (%s)", request_type, url, result, message)

  def anomaly(self):
    return self._test(result_count=3, failure_amount=3)



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise, TestResult
import re
import time
from slapos.networkbench.ping import ping, ping6

@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # set periodicity to run the promise twice per day
    self.custom_frequency = int(self.getConfig('frequency', 720))
    self.setPeriodicity(self.custom_frequency)
    # Skip test check on this promise
    self.setTestLess()

  def sense(self):
    """
      Check re6st optimal status
    """
    # promise ipv6 and ipv4 address to compare.
    ipv4 = self.getConfig('ipv4')
    ipv6 = self.getConfig('ipv6')
    count = int(self.getConfig('count', 10))
    if not ipv4:
      raise ValueError("'ipv4' was not set in promise parameters.")
    if not ipv6:
      raise ValueError("'ipv6' was not set in promise parameters.")

    result_ipv4 = ping(ipv4, count=count)
    result_ipv6 = ping6(ipv6, count=count)
    # push into to the log file
    self.logger.info("%s host=%s code=%s, result=%s, packet_lost_ratio=%s msg=%s" % result_ipv4)
    self.logger.info("%s host=%s code=%s, result=%s, packet_lost_ratio=%s msg=%s" % result_ipv6)

    if result_ipv4[3] == "failed" and result_ipv6[3] != "failed":
      # IPv4 is unreacheable
      self.logger.info("OK: IPv4 unreachable, IPv6 reachable")
      return

    if result_ipv6[3] == "failed":
      # IPv6 is unreacheable
      self.logger.error("FAILED: IPv4 reachable, IPv6 unreachable")
      return

    latency4 = float(result_ipv4[3])
    latency6 = float(result_ipv6[3])
    # We can consider that at worst 1ms is added to
    # ipv4 response, due the usage of openvpn.
    acceptable_delay = int(self.getConfig('acceptable-delay', 1))
    # We can consider that we accept a certain increase
    # on latency, if we are on a bit congested link.
    # So 10% is reseonable enough.
    acceptable_lost = int(self.getConfig('acceptable-lost', 0.10))
    # Increase latency with the value.
    latency4 += acceptable_delay + latency4 * acceptable_lost
    if latency4 < latency6:
      self.logger.error("FAIL %s (latency4) > %s (latence6)" % (latency4, latency6))
    else:
      # Compare if both has Same working rate
      self.logger.info("OK: IPv4 reachable, IPv6 reachable")

  def anomaly(self):
    # only check the result of the two latest sense call
    return self._test(result_count=2, failure_amount=2, latest_minute=self.custom_frequency*3)



#!/usr/bin/env python

from __future__ import print_function

import sys
import os
import glob
import json
from six.moves import configparser
import time
from datetime import datetime
import base64
import hashlib
import PyRSS2Gen

from slapos.util import bytes2str, str2bytes

def getKey(item):
  return item.source.name

class MonitorFeed(object):

  def __init__(self, instance_name, hosting_name,
      public_url, private_url, feed_url):
    self.rss_item_list = []
    self.report_date = datetime.utcnow()
    self.instance_name = instance_name
    self.hosting_name = hosting_name
    self.public_url = public_url
    self.private_url = private_url
    self.feed_url = feed_url

  def appendItem(self, item_dict, has_string=""):
    event_date = item_dict['result']['change-date']
    report_date = item_dict['result']['date']
    description = item_dict['result'].get('message', '')
    guid = base64.b64encode(str2bytes("%s, %s, %s, %s" % (self.hosting_name,
      item_dict['title'], has_string, event_date)))
    rss_item = PyRSS2Gen.RSSItem(
      categories = [item_dict['status']],
      source = PyRSS2Gen.Source(item_dict['title'], self.public_url),
      title = '[%s] %s' % (item_dict['status'], item_dict['title']),
      description = "\n%s" % (description,),
      link = self.private_url,
      pubDate = event_date,
      guid = PyRSS2Gen.Guid(bytes2str(guid), isPermaLink=False)
    )
    self.rss_item_list.append(rss_item)

  def generateRSS(self, output_file):
    ### Build the rss feed
    # try to keep the list in the same order
    sorted(self.rss_item_list, key=getKey)
    rss_feed = PyRSS2Gen.RSS2 (
      title = self.instance_name,
      link = self.feed_url,
      description = self.hosting_name,
      lastBuildDate = self.report_date,
      items = self.rss_item_list
    )

    with open(output_file, 'w') as frss:
      frss.write(rss_feed.to_xml())

def generateStatisticsData(stat_file_path, content):
  # csv document for success/error statictics
  if not os.path.exists(stat_file_path) or os.stat(stat_file_path).st_size == 0:
    with open(stat_file_path, 'w') as fstat:
      data_dict = {
        "date": time.time(),
        "data": ["Date, Success, Error, Warning"]
      }
      fstat.write(json.dumps(data_dict))

  current_state = ''
  if 'state' in content:
    current_state = '%s, %s, %s, %s' % (
      content['date'],
      content['state']['success'],
      content['state']['error'],
      '')

  # append to file
  if current_state:
    with open (stat_file_path, mode="r+") as fstat:
      fstat.seek(0,2)
      position = fstat.tell() -2
      fstat.seek(position)
      fstat.write('%s}' % ',"{}"]'.format(current_state))

def writeDocumentList(folder_path):
  # Save document list in a file called _document_list
  public_document_list = [os.path.splitext(file)[0]
                for file in os.listdir(folder_path) if file.endswith('.json')]

  with open(os.path.join(folder_path, '_document_list'), 'w') as lfile:
    lfile.write('\n'.join(public_document_list))

def generateMonitoringData(config, public_folder, private_folder, public_url,
    private_url, feed_url):
  feed_output = os.path.join(public_folder, 'feed')
  # search for all status files
  file_list = list(filter(
    os.path.isfile,
    glob.glob("%s/promise/*.status.json" % public_folder)
  ))

  promises_status_file = os.path.join(private_folder, '_promise_status')
  previous_state_dict = {}
  new_state_dict = {}
  error = success = 0
  monitor_feed = MonitorFeed(
    config.get('monitor', 'title'),
    config.get('monitor', 'root-title'),
    public_url,
    private_url,
    feed_url)

  if os.path.exists(promises_status_file):
    with open(promises_status_file) as f:
      try:
        previous_state_dict = json.loads(f.read())
      except ValueError:
        pass

  # clean up stale history files
  expected_history_json_name_list = [
    os.path.basename(q).replace('status.json', 'history.json') for q in file_list]
  cleanup_history_json_path_list = []
  for history_json_name in [q for q in os.listdir(public_folder) if q.endswith('history.json')]:
    if history_json_name not in expected_history_json_name_list:
      cleanup_history_json_path_list.append(os.path.join(public_folder, history_json_name))
  for cleanup_path in cleanup_history_json_path_list:
    try:
      os.unlink(cleanup_path)
    except Exception:
      print('ERROR: Failed to remove stale %s' % (cleanup_path,))
    else:
      print('OK: Removed stale %s' % (cleanup_path,))

  for file in file_list:
    try:
      with open(file, 'r') as temp_file:
        tmp_json = json.loads(temp_file.read())

      if tmp_json['result']['failed']:
        promise_status = "ERROR"
        error += 1
      else:
        promise_status = "OK"
        success += 1
      tmp_json['result']['change-date'] = tmp_json['result']['date']
      if tmp_json['name'] in previous_state_dict:
        status, change_date, _ = previous_state_dict[tmp_json['name']]
        if promise_status == status:
          tmp_json['result']['change-date'] = change_date

      tmp_json['status'] = promise_status
      message_hash = hashlib.md5(
        str2bytes(tmp_json['result'].get('message', ''))).hexdigest()
      new_state_dict[tmp_json['name']] = [
        promise_status,
        tmp_json['result']['change-date'],
        message_hash
      ]
      monitor_feed.appendItem(tmp_json, message_hash)
      savePromiseHistory(
        tmp_json['title'],
        tmp_json,
        previous_state_dict.get(tmp_json['name']),
        public_folder
      )
    except ValueError as e:
      # bad json file
      print("ERROR: Bad json file at: %s\n%s" % (file, e))
      continue

  with open(promises_status_file, "w") as f:
    json.dump(new_state_dict, f)

  monitor_feed.generateRSS(feed_output)
  return error, success

def savePromiseHistory(promise_name, state_dict, previous_state_list,
    history_folder):
  if not os.path.exists(history_folder) and os.path.isdir(history_folder):
    self.logger.warning('Bad promise history folder, history is not saved...')
    return

  history_file = os.path.join(
    history_folder,
    '%s.history.json' % promise_name
  )

  # Remove useless informations
  result = state_dict.pop('result')
  state_dict.update(result)
  state_dict.pop('path', '')
  state_dict.pop('type', '')
  if not os.path.exists(history_file) or not os.stat(history_file).st_size:
    with open(history_file, 'w') as f:
      data_dict = {
        "date": time.time(),
        "data": [state_dict]
      }
      json.dump(data_dict, f)
  else:
    if previous_state_list is not None:
      _, change_date, checksum = previous_state_list
      current_sum = hashlib.md5(str2bytes(state_dict.get('message', ''))).hexdigest()
      if state_dict['change-date'] == change_date and \
          current_sum == checksum:
        # Only save the changes and not the same info
        return

    state_dict.pop('title', '')
    state_dict.pop('name', '')
    with open (history_file, mode="r+") as f:
      f.seek(0,2)
      f.seek(f.tell() -2)
      f.write('%s}' % ',{}]'.format(json.dumps(state_dict)))

def run(monitor_conf_file):

  config = configparser.ConfigParser()
  config.read(monitor_conf_file)

  base_folder = config.get('monitor', 'private-folder')
  status_folder = config.get('monitor', 'public-folder')
  base_url = config.get('monitor', 'base-url')
  related_monitor_list = config.get("monitor", "monitor-url-list").split()
  statistic_folder = os.path.join(base_folder, 'documents')
  # need webdav to update parameters
  parameter_file = os.path.join(base_folder, 'config', '.jio_documents', 'config.json')

  public_url = "%s/share/public/" % base_url
  private_url = "%s/share/private/" % base_url
  feed_url = "%s/public/feed" % base_url
  status = 'OK'
  global_state_file = os.path.join(base_folder, 'monitor.global.json')
  public_state_file = os.path.join(status_folder, 'monitor.global.json')
  report_date = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S+0000')

  error, success = generateMonitoringData(config, status_folder, base_folder,
                                          public_url, private_url, feed_url)
  if error:
    status = 'ERROR'

  global_state_dict = dict(
    status=status,
    state={
      'error': error,
      'success': success
    },
    type='global', # bwd compatibility
    portal_type='Software Instance',
    date=report_date,
    _links={"rss_url": {"href": feed_url},
            "public_url": {"href": public_url},
            "private_url": {"href": private_url},
            "related_monitor": []
          },
    data={'state': 'monitor_state.data',
          'process_state': 'monitor_process_resource.status',
          'process_resource': 'monitor_resource_process.data',
          'memory_resource': 'monitor_resource_memory.data',
          'io_resource': 'monitor_resource_io.data',
          'monitor_process_state': 'monitor_resource.status'},
    title=config.get('monitor', 'title'),
    specialise_title=config.get('monitor', 'root-title'),
    aggregate_reference=config.get('promises', 'computer-id'),
    ipv4=config.get('promises', 'ipv4'),
    ipv6=config.get('promises', 'ipv6'),
    software_release=config.get('promises', 'software-release'),
    software_type=config.get('promises', 'software-type'),
    partition_id=config.get('promises', 'partition-id'),
  )

  if not global_state_dict['title']:
    global_state_dict['title'] = 'Instance Monitoring'

  if related_monitor_list:
    global_state_dict['_links']['related_monitor'] = [{'href': "%s/share/public" % url}
                          for url in related_monitor_list]

  if os.path.exists(parameter_file):
    with open(parameter_file) as cfile:
      global_state_dict['parameters'] = json.loads(cfile.read())

  # Public information with the link to private folder
  public_state_dict = dict(
    status=status,
    date=report_date,
    _links={'monitor': {'href': '%s/share/private/' % base_url}},
    title=global_state_dict.get('title', ''),
    specialise_title=global_state_dict.get('specialise_title', ''),
  )
  public_state_dict['_links']['related_monitor'] = global_state_dict['_links'].get('related_monitor', [])

  with open(global_state_file, 'w') as fglobal:
    fglobal.write(json.dumps(global_state_dict))

  with open(public_state_file, 'w') as fpglobal:
    fpglobal.write(json.dumps(public_state_dict))

  # write list of files
  writeDocumentList(status_folder)
  writeDocumentList(base_folder)
  writeDocumentList(statistic_folder)

  generateStatisticsData(
    os.path.join(statistic_folder, 'monitor_state.data.json'),
    global_state_dict
  )

  return 0

def main():
  if len(sys.argv) < 2:
    print("Usage: %s <monitor_conf_path>" % sys.argv[0])
    sys.exit(2)
  sys.exit(run(sys.argv[1]))



##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from slapos.test.promise.plugin import TestPromisePluginMixin
from slapos.grid.promise import PromiseError
import os

class TestCheckServerCPULoad(TestPromisePluginMixin):

  def setUp(self):
    super(TestCheckServerCPULoad, self).setUp()
    self.promise_name = "server-cpu-load-promise.py"

    content = """from slapos.promise.plugin.check_server_cpu_load import RunPromise

extra_config_dict = {
  'cpu-load-threshold': '2.0',
}
"""
    self.writePromise(self.promise_name, content)

  def test_check_cpu_load_run(self):
    self.configureLauncher(timeout=5)
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    if result['result']['failed']:
      self.assertTrue("CPU load is high" in result['result']['message'])
    else:
      self.assertEqual("CPU load is OK", result['result']['message'].strip())



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise, TestResult
import re
import time
from slapos.networkbench.ping import ping, ping6

@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # set periodicity to run the promise twice per day
    self.custom_frequency = int(self.getConfig('frequency', 720))
    self.setPeriodicity(self.custom_frequency)
    # Skip test check on this promise
    self.setTestLess()

  def sense(self):
    """
      Check if there ICMP packets lost on given address
    """
    # Address to ping to
    address = self.getConfig('address')
    if not address:
      raise ValueError("'address' was not set in promise parameters.")
    # Force use ipv4 protocol ?
    ipv4 = self.getConfig('ipv4') in ('True', 'true', '1')
    count = int(self.getConfig('count', 10))
    threshold = int(self.getConfig('threshold', 0))
    if threshold < 0:
      raise ValueError("'threshold' value should be greater than 0.")

    if ipv4:
      result = ping(address, count=count)
    else:
      result = ping6(address, count=count)

    message = "%s host=%s code=%s, result=%s, packet_lost_ratio=%s msg=%s" % result
    packet_lost_ratio = int(result[4])
    if packet_lost_ratio == -1 or packet_lost_ratio > threshold:
      # Packet lost occurred
      self.logger.error(message)
    else:
      self.logger.info(message)

  def anomaly(self):
    # only check the result of the two latest sense call
    return self._test(result_count=2, failure_amount=2, latest_minute=self.custom_frequency*3)



from __future__ import division, print_function

import argparse
import itertools
import os
import re
import shutil
import subprocess
import sys
import time

from datetime import datetime
from .runner_utils import *
from six.moves import map

os.environ['LC_ALL'] = 'C'
os.umask(0o77)


def parseArgumentList():
  parser = argparse.ArgumentParser()
  parser.add_argument('--srv-path', required=True)
  parser.add_argument('--backup-path', required=True)
  parser.add_argument('--etc-path', required=True)
  parser.add_argument('--rsync-binary', default='rsync')
  parser.add_argument('--backup-wait-time', type=int, required=True)
  parser.add_argument('-n', action='store_true', dest='dry', default=False)

  return parser.parse_args()


def rsync(rsync_binary, source, destination, exclude_list=None, extra_args=None, dry=False):
  arg_list = [
    rsync_binary,
    '-rlptgov',
    '--stats',
    '--safe-links',
    '--ignore-missing-args',
    '--delete',
    '--delete-excluded'
  ]
  if isinstance(exclude_list, list):
    arg_list.extend(["--exclude={}".format(x) for x in sorted(exclude_list)])
  if isinstance(extra_args, list):
    arg_list.extend(extra_args)
  if isinstance(source, list):
    arg_list.extend(sorted(source))
  else:
    arg_list.append(source)
  arg_list.append(destination)

  if dry:
    print('DEBUG:', arg_list)
    return

  try:
    print(subprocess.check_output(arg_list))
  except subprocess.CalledProcessError as e:
    # All rsync errors are not to be considered as errors
    allowed_error_message_list = \
      '^(file has vanished: |rsync warning: some files vanished before they could be transferred)'
    if e.returncode != 24 or \
        re.search(allowed_error_message_regex, e.output, re.M) is None:
      raise


def synchroniseRunnerConfigurationDirectory(config, backup_path):
  if not os.path.exists(backup_path):
    os.makedirs(backup_path)

  file_list = ['config.json']
  # `sorted` is used for Python 2-3 compatibility
  for hidden_file in sorted(os.listdir('.')):
    if hidden_file[0] == '.':
      file_list.append(hidden_file)
  rsync(config.rsync_binary, file_list, backup_path, dry=config.dry)


def synchroniseRunnerWorkingDirectory(config, backup_path):
  file_list = []
  exclude_list = []

  if os.path.isdir('instance'):
    file_list.append('instance')
    # `sorted` is used for Python 2-3 compatibility
    exclude_list = sorted(getExcludePathList(os.getcwd()))

  # XXX: proxy.db should be properly dumped to leverage its
  # atomic properties
  for file in ('project', 'public', 'proxy.db'):
    if os.path.exists(file):
      file_list.append(file)

  if file_list:
    rsync(
      config.rsync_binary, file_list, backup_path,
      exclude_list=exclude_list,
      dry=config.dry
    )


def getBackupFilesModifiedDuringExportList(config, export_start_date):
  export_time = time.time() - export_start_date
  # find all files that were modified during export
  modified_files = subprocess.check_output((
      'find', 'instance', '-cmin',  str(export_time / 60), '-type', 'f', '-path', '*/srv/backup/*'
    ))
  if not modified_files:
    return ()

  # filter those modified files through rsync --exclude getExcludePathList.
  # Indeed, some modified files may be listed in getExcludePathList and in this
  # case, we won't copy them to PBS so it's not really important if they are
  # modified.
  rsync_arg_list = [
    config.rsync_binary,
    '-n',
    '--out-format=%n',
    '--files-from=-',
    '--relative',
    '--no-implied-dirs'
  ]
  rsync_arg_list += map("--exclude={}".format, getExcludePathList(os.getcwd()))
  rsync_arg_list += '.', 'unexisting_dir_or_file_just_to_have_the_output'
  process = subprocess.Popen(rsync_arg_list, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  output = process.communicate(modified_files)[0]
  retcode = process.poll()
  if retcode:
      raise CalledProcessError(retcode, rsync_arg_list[0], output=output)

  important_modified_file_list = output.splitlines()
  not_important_modified_file_set = set(modified_files.splitlines()).difference(important_modified_file_list)
  if not_important_modified_file_set:
    print("WARNING: The following files in srv/backup were modified since the exporter started (srv/backup should contain almost static files):", *sorted(not_important_modified_file_set), sep='\n')

  return important_modified_file_list

def runExport():
  export_start_date = int(time.time())
  print(datetime.fromtimestamp(export_start_date).isoformat())

  args = parseArgumentList()

  def _rsync(*params):
    return rsync(args.rsync_binary, *params, dry=args.dry)

  runner_working_path = os.path.join(args.srv_path, 'runner')
  backup_runner_path = os.path.join(args.backup_path, 'runner')

  # Synchronise runner's etc directory
  with CwdContextManager(args.etc_path):
    with open('.resilient-timestamp', 'w') as f:
      f.write(str(export_start_date))

    # "+ '/'" is mandatory otherwise rsyncing the etc directory
    # will create in the backup_etc_path only a file called etc
    backup_etc_path = os.path.join(args.backup_path, 'etc') + '/'
    synchroniseRunnerConfigurationDirectory(args, backup_etc_path)

  # Synchronise runner's working directory
  # and aggregate signature functions as we are here
  with CwdContextManager(runner_working_path):
    synchroniseRunnerWorkingDirectory(args, backup_runner_path)
    slappart_signature_method_dict = getSlappartSignatureMethodDict()

  # Calculate signature of synchronised files
  with CwdContextManager(args.backup_path):
    writeSignatureFile(slappart_signature_method_dict, runner_working_path)

  # BBB: clean software folder if it was synchronized
  # in an old instance
  backup_software_path = os.path.join(backup_runner_path, 'software')
  if os.path.isdir(backup_software_path):
    shutil.rmtree(backup_software_path)


  # Wait a little to increase the probability to detect an ongoing backup.
  time.sleep(10)

  # Check that export didn't happen during backup of instances
  with CwdContextManager(runner_working_path):
    modified_file_list = getBackupFilesModifiedDuringExportList(args, export_start_date)
    if len(modified_file_list):
      print("ERROR: The following files in srv/backup were modified since the exporter started. Since they must be backup, exporter should be re-run."
            " Let's sleep %s minutes, to let the backup end.\n%s" % (
              args.backup_wait_time, '\n'.join(modified_file_list)))
      time.sleep(args.backup_wait_time * 60)
      sys.exit(1)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import slapos.slap

import glob
import logging
import os
import subprocess
import sys
import time
from six.moves.urllib.request import urlopen

UNIT_TEST_ERP5TESTNODE = 'UnitTest'

class ResiliencyTestSuite(object):
  """
  Abstract class supposed to be extended by Resiliency Test Suites.
  """
  def __init__(self,
               server_url, key_file, cert_file,
               computer_id, partition_id, software,
               namebase,
               root_instance_name,
               sleep_time_between_test=900,
               total_instance_count="2",
               type=None):
    self.server_url = server_url
    self.key_file = key_file
    self.cert_file = cert_file
    self.computer_id = computer_id
    self.partition_id = partition_id
    self.software = software
    self.namebase = namebase
    self.total_instance_count = total_instance_count
    self.root_instance_name = root_instance_name
    self.sleep_time_between_test = sleep_time_between_test
    self.test_type = type

    slap = slapos.slap.slap()
    slap.initializeConnection(server_url, key_file, cert_file)
    self.partition = slap.registerComputerPartition(
        computer_guid=computer_id,
        partition_id=partition_id
    )

    self.logger = logging.getLogger('SlaprunnerResiliencyTest')
    self.logger.setLevel(logging.DEBUG)

  def _doTakeover(self, namebase, target_clone):
    """
    Private method.
    Make the specified clone instance takeover the main instance.
    """
    self.logger.info('Replacing main instance by clone instance %s%s...' % (
        self.namebase, target_clone))

    root_partition_parameter_dict = self._getPartitionParameterDict()
    takeover_url = root_partition_parameter_dict['takeover-%s-%s-url' % (namebase, target_clone)]
    takeover_password = root_partition_parameter_dict['takeover-%s-%s-password' % (namebase, target_clone)]

    # Do takeover
    takeover_result = urlopen('%s?password=%s' % (takeover_url, takeover_password)).read()
    if 'Error' in takeover_result:
      raise Exception('Error while doing takeover: %s' % takeover_result)

    self.logger.info('Done.')

  def generateData(self):
    """
    Generate data that will be used by the test.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def pushDataOnMainInstance(self):
    """
    Push our data to the main instance.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def checkDataOnCloneInstance(self):
    """
    Check that, on the ex-clone, now-main instance, data is the same as
    what we pushed to the ex-main instance.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def deleteTimestamp():
    """
    XXX-Nicolas delete .timestamp in test partition to force the full processing
    by slapgrid, to force the good parameters to be passed to the instances of the tree
    """
    home = os.getenv('HOME')
    timestamp = os.path.join(home, '.timestamp')
    os.remove(timestamp)

  def _getPartitionParameterDict(self):
    """
    Helper.
    Return the partition parameter dict of the main root ("resilient") instance.
    """
    return self.partition.request(
        software_release=self.software,
        software_type='resilient',
        partition_reference=self.root_instance_name
    ).getConnectionParameterDict()
    self.deleteTimestamp()

  def _returnNewInstanceParameter(self, parameter_key, old_parameter_value, force_new=False):
    """
    Helper, can be used inside of checkDataOnCloneInstance.
    Wait for the new parameter (of old-clone new-main instance) to appear.
    Check than it is different from the old parameter
    """
    # if we are inside of a classical erp5testnode: just return the same parameter.
    if self.test_type == UNIT_TEST_ERP5TESTNODE and not force_new:
      return old_parameter_value

    self.logger.info('Waiting for new main instance to be ready...')
    new_parameter_value = None
    while not new_parameter_value or new_parameter_value == 'None' or  new_parameter_value == old_parameter_value:
      self.logger.info('Not ready yet. SlapOS says new parameter value is %s' % new_parameter_value)
      new_parameter_value = self._getPartitionParameterDict().get(parameter_key, None)
      time.sleep(30)
    self.logger.info('New parameter value of instance is %s' % new_parameter_value)
    return new_parameter_value

  def _waitForCloneToBeReadyForTakeover(self, clone):
    self.logger.info('Wait for Clone %s to be ready for takeover' % clone)

    root_partition_parameter_dict = self._getPartitionParameterDict()
    takeover_url = root_partition_parameter_dict['takeover-%s-%s-url' % (self.namebase, clone)]
    takeover_password = root_partition_parameter_dict['takeover-%s-%s-password' % (self.namebase, clone)]

    # Connect to takeover web interface and wait for importer script to be not running
    takeover_page_content = urlopen(takeover_url).read()
    while "<b>Importer script(s) of backup in progress:</b> True" in takeover_page_content:
      time.sleep(10)
      takeover_page_content = urlopen(takeover_url).read()
    return

  def _testClone(self, clone):
    """
    Private method.
    Launch takeover and check for a specific clone.
    """
    # Wait for XX minutes so that replication is done
    self.logger.info(
      'Sleeping for %s seconds before testing clone %s.' % (
          self.sleep_time_between_test,
          clone
        ))
    time.sleep(self.sleep_time_between_test)

    self._waitForCloneToBeReadyForTakeover(clone)

    # Before doing takeover we expect the instances to be in a stable state
    if not self._testPromises():
      return False

    self.logger.info('Testing %s%s instance.' % (self.namebase, clone))
    self._doTakeover(self.namebase, clone)

    if self.test_type == UNIT_TEST_ERP5TESTNODE: # Run by classical erp5testnode using slapproxy
      # Run manually slapos node instance
      # XXX hardcoded path
      self.logger.info('Running "slapos node instance"...')
      slapos_configuration_file_path = os.path.join(
          os.path.dirname(sys.argv[0]),
          '..', '..', '..', 'slapos.cfg'
      )
      # Output is huge and we don't want to store it in memory nor print it
      devnull = open('/dev/null', 'w')
      command = [os.path.join(os.environ['HOME'], 'software_release', 'bin', 'slapos'), 'node', 'instance',
                 '--cfg=%s' % slapos_configuration_file_path,
                 '--pidfile=slapos.pid']
      for _ in range(5):
        subprocess.Popen(command, stdout=devnull, stderr=devnull).wait()

    success = self.checkDataOnCloneInstance()

    if success:
      return True

  def _testPromises(self):
    """
    Run promises in all instances (export, PBS, import(s)) and
    check their output. An error at any step of the resilience
    should make at least one of the promises complain.
    """
    for promise_directory in glob.glob(
        os.path.realpath(os.path.join(os.path.dirname(sys.argv[0]), '..', '..', '..', 'inst*', '*', 'etc', 'promise'))):
      self.logger.info("Testing promises of instance's directory : %s", promise_directory)
      promise_list = os.listdir(promise_directory)
      for promise in promise_list:
        # XXX: for the moment ignore monitor promises as they can never succeed
        # in a testnode environment.
        if 'monitor' in promise:
          continue
        try:
          subprocess.check_output(os.path.join(promise_directory, promise),
                                  stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
          self.logger.error('ERROR : promise "%s" failed with output :\n%s', promise, e.output)
          return False

    return True

  def runTestSuite(self):
    """
    Generate data to send,
    Push data on main instance,
    Wait for replication to be done,
    For each clone: Do a takeover, Check data.
    """
    self.generateData()

    self.pushDataOnMainInstance()

    # In resilient stack, main instance (example with KVM) is named "kvm0",
    # clones are named "kvm1", "kvm2", ...
    clone_count = int(self.total_instance_count) - 1
    # So first clone starts from 1.
    current_clone = 1

    # In case we have only one clone: test the takeover twice
    # so that we test the reconstruction of a new clone.
    if clone_count == 1:
      return self._testClone(1)
      #for i in range(2):
      #  success = self._testClone(1)
      #  if not success:
      #    return False

    else:
      # Test each clone
      while current_clone <= clone_count:
        success = self._testClone(current_clone)
        if not success:
          return False
        current_clone = current_clone + 1

    # All clones have been successfully tested: success.
    return True






def tail_file(file_path, line_count=10):
  """
  Returns the last lines of file.
  """

  line_list = []
  with open(file_path) as f:
    BUFSIZ = 1024
    f.seek(0, 2)
    bytes = f.tell()
    size = line_count + 1
    block = -1
    while size > 0 and bytes > 0:
      if bytes - BUFSIZ > 0:
          # Seek back one whole BUFSIZ
          f.seek(block * BUFSIZ, 2)
          line_list.insert(0, f.read(BUFSIZ))
      else:
          f.seek(0, 0)
          # only read what was not read
          line_list.insert(0, f.read(bytes))
      line_len = line_list[0].count('\n')
      size -= line_len
      bytes -= BUFSIZ
      block -= 1

  return '\n'.join(''.join(line_list).splitlines()[-line_count:])


##############################################################################
#
# Copyright (c) 2017 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import unittest
import os.path
import socket
import time
import psutil
from slapos.promise.apache_mpm_watchdog import watchServerStatus, \
  loadJSONFile, writeJSONFile, getServerStatus, search_pid_regex

from . import data

class TestApacheMPMWatchdog(unittest.TestCase):

  base_path, = data.__path__

  def text_searchPidRegex(self):

    with open(self.base_path + "/server_status.html") as f:
      server_status = f.read()
      f.close()

    self.assertEqual(['12345', '12346'], 
      re.findall(search_pid_regex, server_status))

    

  def test_loadJSONFile(self):
    self.assertEqual({},
       loadJSONFile("couscous"))

    self.assertEqual(
      {"1234": 1496161635.514768 , "4321": 1496161635.514768},
      loadJSONFile(os.path.join(self.base_path, "test_db.json")))

    self.assertEqual(
      {},
      loadJSONFile(os.path.join(self.base_path, "corrupted_db.json")))

  def test_writeJSONFile(self):
    # Check if don't raise.
    self.assertEqual(None,
      writeJSONFile({}, None))

    current_pid = os.getpid() 
    self.assertEqual(None,
      writeJSONFile({"123482": 123, current_pid: 124},
          os.path.join(self.base_path, "write_db.json")))

    with open(os.path.join(self.base_path, "write_db.json")) as f:
      json_content = f.read()
      f.close()

    self.assertEqual(json_content,
      '{"%s": 124}' % current_pid)


  def test_getServerStatus(self):
    self.assertEqual(None,
        getServerStatus("http://localhost/", None, None))
    self.assertEqual(None,
        getServerStatus("http://localhost/", 
                           "user", "password"))
    self.assertNotEqual(None,
        getServerStatus("https://www.erp5.com/", None, None))


if __name__ == '__main__':
  unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import sys
try:
  import errno
except ImportError:
  from os import errno
import os
import stat
import json
from six.moves import configparser
import traceback
import argparse
from six.moves.urllib.request import urlopen
from six.moves.urllib.error import HTTPError
import ssl
import glob
import socket
from datetime import datetime
from xml.sax.saxutils import escape

from slapos.util import bytes2str

OPML_START = """<?xml version="1.0" encoding="UTF-8"?>
<!-- OPML generated by SlapOS -->
<opml version="1.1">
	<head>
		<title>%(root_title)s</title>
		<dateCreated>%(creation_date)s</dateCreated>
		<dateModified>%(modification_date)s</dateModified>
	</head>
	<body>
	  <outline text="%(outline_title)s">"""
OPML_END = """	  </outline>
  </body>
</opml>"""

OPML_OUTLINE_FEED = '<outline text="%(title)s" title="%(title)s" type="rss" version="RSS" htmlUrl="%(html_url)s" xmlUrl="%(xml_url)s" url="%(global_url)s" />'


def parseArguments():
  """
  Parse arguments for monitor instance.
  """
  parser = argparse.ArgumentParser()
  parser.add_argument('-c', '--config-file', required=True,
                      default='monitor.cfg',
                      help='Monitor Configuration file')

  return parser.parse_args()


def mkdirAll(path):
  try:
    os.makedirs(path)
  except OSError as e:
    if e.errno == errno.EEXIST and os.path.isdir(path):
      pass
    else: raise

def softConfigGet(config, *args, **kwargs):
  try:
    return config.get(*args, **kwargs)
  except (configparser.NoOptionError, configparser.NoSectionError):
    return None

def createSymlink(source, destination):
  try:
    os.symlink(source, destination)
  except OSError as e:
    if e.errno != errno.EEXIST:
      raise

class Monitoring(object):

  def __init__(self, configuration_file):
    self._config_file = configuration_file
    config = self.loadConfig([configuration_file])

    # Set Monitor variables
    self.title = config.get("monitor", "title")
    self.root_title = config.get("monitor", "root-title")
    self.service_pid_folder = config.get("monitor", "service-pid-folder")
    self.crond_folder = config.get("monitor", "crond-folder")
    self.public_folder = config.get("monitor", "public-folder")
    self.private_folder = config.get("monitor", "private-folder")
    self.webdav_folder = config.get("monitor", "webdav-folder")
    self.webdav_url = '%s/share' % config.get("monitor", "base-url")
    self.public_url = '%s/public' % config.get("monitor", "base-url")
    self.public_path_list = config.get("monitor", "public-path-list").split()
    self.private_path_list = config.get("monitor", "private-path-list").split()
    self.monitor_url_list = config.get("monitor", "monitor-url-list").split()
    self.parameter_list = [param.strip() for param in config.get("monitor", "parameter-list").split('\n') if param]
    # Use this file to write knowledge0_cfg required by webrunner
    self.parameter_cfg_file = config.get("monitor", "parameter-file-path").strip()
    self.pid_file = config.get("monitor", "pid-file")
    self.promise_output_file = config.get("monitor", "promise-output-file")
    self.promise_folder = config.get("promises", 'promise-folder')
    if config.has_option("promises", 'legacy-promise-folder'):
      self.legacy_promise_folder = config.get("promises", 'legacy-promise-folder')
    else:
      self.legacy_promise_folder = None
    self.promise_output = config.get("promises", 'output-folder')

    self.config_folder = os.path.join(self.private_folder, 'config')
    self.data_folder = config.get("monitor", "document-folder")
    self.bootstrap_is_ok = True

  def loadConfig(self, pathes, config=None):
    if config is None:
      config = configparser.ConfigParser()
    try:
      config.read(pathes)
    except configparser.MissingSectionHeaderError:
      traceback.print_exc()
    return config

  def readInstanceConfiguration(self):
    type_list = ['raw', 'file', 'htpasswd', 'httpdcors']
    configuration_list = []

    if not self.parameter_list:
      return []
  
    for config in self.parameter_list:
      config_list = config.strip().split(' ')
      # type: config_list[0]
      if len(config_list) >= 3 and config_list[0] in type_list:
        if config_list[0] == 'raw':
          configuration_list.append(dict(
            key='',
            title=config_list[1],
            value=' '.join(config_list[2:])
          ))
        elif (config_list[0] == 'file' or config_list[0] == 'htpasswd'):
          directory = os.path.dirname(config_list[2])
          if not os.path.exists(directory) or not os.access(directory, os.W_OK):
            raise OSError("Directory does not exists or does not have write acess")
          if os.path.exists(config_list[2]) and os.path.isfile(config_list[2]):
            try:
              with open(config_list[2]) as cfile:
                param_value = cfile.read()
            except OSError as e:
              print('Cannot read file %s, Error is: %s' % (config_list[2], e))
              pass
          else:
            param_value = ""
          parameter = dict(
            key=config_list[1],
            title=config_list[1],
            value=param_value,
            description={
              "type": config_list[0],
              "file": config_list[2]
            }
          )
          if config_list[0] == 'htpasswd':
            if len(config_list) != 5 or not os.path.exists(config_list[4]):
              print('htpasswd file is not specified: %s' % config_list)
              continue
            parameter['description']['user'] = config_list[3]
            parameter['description']['htpasswd'] = config_list[4]
          configuration_list.append(parameter)
        elif config_list[0] == 'httpdcors' and os.path.exists(config_list[2]) and \
            os.path.exists(config_list[3]):
          old_cors_file = os.path.join(
            os.path.dirname(config_list[2]),
            'prev_%s' % os.path.basename(config_list[2])
          )
          try:
            cors_content = ""
            if os.path.exists(old_cors_file):
              with open(old_cors_file) as cfile:
                cors_content = cfile.read()
            else:
              # Create empty file
              with open(old_cors_file, 'w') as cfile:
                cfile.write("")
            parameter = dict(
              key=config_list[1],
              title=config_list[1],
              value=cors_content,
              description={
                "type": config_list[0],
                "cors_file": config_list[2],
                "gracefull_bin": config_list[3]
              }
            )
            configuration_list.append(parameter)
          except OSError as e:
            print('Cannot read file at %s, Error is: %s' % (old_cors_file, e))
            pass
    return configuration_list

  def createSymlinksFromConfig(self, destination_folder, source_path_list, name=""):
    if destination_folder:
      if source_path_list:
        for path in source_path_list:
          path = path.rstrip('/')
          dirname = os.path.join(destination_folder, name)
          try:
            mkdirAll(dirname)  # could also raise OSError
            os.symlink(path, os.path.join(dirname, os.path.basename(path)))
          except OSError as e:
            if e.errno != errno.EEXIST:
              raise

  def getMonitorTitleFromUrl(self, monitor_url):
    # This file should be generated
    if not monitor_url.startswith('https://') and not monitor_url.startswith('http://'):
      return 'Unknown Instance'
    if not monitor_url.endswith('/'):
      monitor_url = monitor_url + '/'

    url  = monitor_url + '/monitor.global.json'
    success = False
    monitor_title = 'Unknown Instance'
    try:
      # Timeout after 20 seconds
      timeout = 20
      # XXX - working here with public url
      if hasattr(ssl, '_create_unverified_context'):
        context = ssl._create_unverified_context()
        response = urlopen(url, context=context, timeout=timeout)
      else:
        response = urlopen(url, timeout=timeout)
    except HTTPError:
      print("ERROR: Failed to get Monitor configuration file at %s " % url)
    except (socket.timeout, ssl.SSLError) as e:
      print("ERROR: Timeout with %r while downloading monitor config at %s " % (e, url))
    else:
      try:
        monitor_dict = json.loads(bytes2str(response.read()))
        monitor_title = monitor_dict.get('title', 'Unknown Instance')
        success = True
      except ValueError as e:
        print("ERROR: Json file at %s is not valid" % url)

    self.bootstrap_is_ok = success
    return monitor_title

  def configureFolders(self):
    # create symlinks from monitor.conf
    self.createSymlinksFromConfig(self.public_folder, self.public_path_list)
    self.createSymlinksFromConfig(self.private_folder, self.private_path_list)
    # configure public and private folder
    self.createSymlinksFromConfig(self.webdav_folder, [self.public_folder])
    self.createSymlinksFromConfig(self.webdav_folder, [self.private_folder])

    config_jio_folder = os.path.join(self.config_folder, '.jio_documents')
    mkdirAll(config_jio_folder)

  def makeConfigurationFiles(self):
    config_folder = os.path.join(self.config_folder, '.jio_documents')
    parameter_config_file = os.path.join(config_folder, 'config.parameters.json')
    parameter_file = os.path.join(config_folder, 'config.json')

    parameter_list = self.readInstanceConfiguration()
    description_dict = {}

    if parameter_list:
      for i in range(0, len(parameter_list)):
        key = parameter_list[i]['key']
        if key:
          description_dict[key] = parameter_list[i].pop('description')

    with open(parameter_config_file, 'w') as config_file:
      config_file.write(json.dumps(description_dict))

    with open(parameter_file, 'w') as config_file:
      config_file.write(json.dumps(parameter_list))

    try:
      with open(self.parameter_cfg_file, 'w') as pfile:
        pfile.write('[public]\n')
        for parameter in parameter_list:
          if parameter['key']:
            pfile.write('%s = %s\n' % (parameter['key'], parameter['value']))
    except OSError as e:
      print("Error failed to create file %s" % self.parameter_cfg_file)
      pass


  def generateOpmlFile(self, feed_url_list, output_file):

    if os.path.exists(output_file):
      creation_date = datetime.utcfromtimestamp(os.path.getctime(output_file))\
        .strftime("%a, %d %b %Y %H:%M:%S +0000")
      modification_date = datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S +0000")
    else:
      creation_date = modification_date = datetime.utcnow()\
        .strftime("%a, %d %b %Y %H:%M:%S +0000")

    opml_content = OPML_START % {'creation_date': creation_date,
                                  'modification_date': modification_date,
                                  'outline_title': 'Monitoring RSS Feed list',
                                  'root_title': escape(self.root_title)}

    opml_content += OPML_OUTLINE_FEED % {'title': escape(self.title),
        'html_url': self.public_url + '/feed',
        'xml_url': self.public_url + '/feed',
        'global_url': "%s/private/" % self.webdav_url}
    for feed_url in feed_url_list:
      opml_content += OPML_OUTLINE_FEED % {
        'title': self.getMonitorTitleFromUrl(feed_url + "/public"),
        'html_url': feed_url + '/public/feed',
        'xml_url': feed_url + '/public/feed',
        'global_url': "%s/share/private/" % feed_url}

    opml_content += OPML_END

    with open(output_file, 'w') as wfile:
      wfile.write(opml_content)

  def cleanupMonitorDeprecated(self):
    # Monitor report feature is removed
    cleanup_file_list = glob.glob("%s/*.history.json" % self.private_folder)
    cleanup_file_list.extend(glob.glob("%s/*.report.json" % self.private_folder))
    cleanup_file_list.extend(glob.glob("%s/*.report.json" % self.data_folder))
    cleanup_file_list.extend(glob.glob("%s/*.history.json" % self.data_folder))
    cleanup_file_list.extend(glob.glob("%s/*.status.json" % self.public_folder))
    cleanup_file_list.append(self.crond_folder + "/monitor-reports")
    cleanup_file_list.append(self.crond_folder + "/monitor-promises")

    for file in cleanup_file_list:
      try:
        if os.path.exists(file):
          os.unlink(file)
      except OSError as e:
        print("failed to remove file %s." % file, e)

    # cleanup result of promises that was removed
    if self.legacy_promise_folder is not None:
      promise_list = os.listdir(self.legacy_promise_folder)
    else:
      promise_list = []
    for name in os.listdir(self.promise_folder):
      if name.startswith('__init__'):
        continue
      promise_list.append(os.path.splitext(name)[0])
    for name in os.listdir(self.promise_output):
      if not name.endswith('.status.json'):
        continue
      try:
        position = promise_list.index(name.replace('.status.json', ''))
      except ValueError:
        status_path = os.path.join(self.promise_output, name)
        if os.path.exists(status_path):
          try:
            os.unlink(status_path)
          except OSError as e:
            print("Error: Failed to delete %s" % status_path, e)
      else:
        promise_list.pop(position)

  def bootstrapMonitor(self):

    # save pid of current process into file
    with open(self.pid_file, 'w') as pid_file:
      pid_file.write(str(os.getpid()))

    self.configureFolders()

    # Generate OPML file
    self.generateOpmlFile(self.monitor_url_list,
      os.path.join(self.public_folder, 'feeds'))

    # cleanup deprecated entries
    self.cleanupMonitorDeprecated()

    # Generate parameters files and scripts
    self.makeConfigurationFiles()

    # Write an empty file when monitor bootstrap went until the end
    if self.bootstrap_is_ok:
      with open(self.promise_output_file, 'w') as promise_file:
        promise_file.write("")
      print("SUCCESS: bootstrap is OK")
    elif os.path.exists(self.promise_output_file):
      os.unlink(self.promise_output_file)


    return 0


def main():
  parser = parseArguments()

  monitor = Monitoring(parser.config_file)
  
  sys.exit(monitor.bootstrapMonitor())



##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from __future__ import print_function

import argparse
import json
import os
import pprint
import socket
import time
import errno
import psutil
from operator import itemgetter
from slapos.util import str2bytes

def parseArgument():
  """
  Very basic argument parser. Might blow up for anything else than
  "./executable mysocket.sock stop/resume".
  """
  parser = argparse.ArgumentParser()
  parser.add_argument('--suspend', action='store_const', dest='action', const='suspend')
  parser.add_argument('--resume', action='store_const', dest='action', const='resume')
  parser.add_argument('--create-snapshot', action='store_const', dest='action', const='createSnapshot')
  parser.add_argument('--create-internal-snapshot', action='store_const', dest='action', const='createInternalSnapshot')
  parser.add_argument('--delete-internal-snapshot', action='store_const', dest='action', const='deleteInternalSnapshot')
  parser.add_argument('--drive-backup', action='store_const', dest='action', const='driveBackup') 
  parser.add_argument('--query-commands', action='store_const', dest='action', const='queryCommands')
  parser.add_argument('--query-item', dest='query', choices=[
    "cpus", "hotpluggable-cpus", "memory-devices", "memdev", "balloon",
    "pci", "status", "acpi-ospm-status"])
  parser.add_argument('--update-device', action='store_const', dest='action', const='updateDevice') 
  parser.add_argument('--device-options', dest='device_options',
    help="Option used for update-device:\n" \
      '"device=cpu,amount=VALUE_INT[,model=MODEL]"\n OR '\
      '"device=memory,mem=VALUE_MB[,slot=VALUE_MB],nslot=VALUE_INT[,canreboot=0|1]"')

  parser.add_argument('--socket', dest='unix_socket_location', required=True)
  parser.add_argument('remainding_argument_list', nargs=argparse.REMAINDER)
  return  parser.parse_args()

def getInitialQemuResourceDict(pid_file):
  """
    Return CPU ad RAM initial values used to start Qemu process
  """
  if not os.path.exists(pid_file):
    return None

  with open(pid_file) as f:
    try:
      pid = int(f.read())
    except ValueError as e:
      raise ValueError("%r is empty or doesn't contain a valid pid number: %s" % (
        pid_file, str(e)))

  wait_count = 5
  while wait_count > 0:
    try:
      process = psutil.Process(pid)
      break
    except psutil.NoSuchProcess:
      print("Qemu process is not started yet...")
      wait_count -= 1
      time.sleep(0.5)
  else:
    raise ValueError('No process with pid %s' % pid)

  if not process.name().startswith('qemu-system'):
    raise ValueError("The pid %s is used by %r and not a qemu process." % (
      pid, process.name()))
  resource_dict = {'cpu': None, 'ram': None}
  cmd_list = process.cmdline()
  cpu_index = cmd_list.index('-smp')
  ram_index = cmd_list.index('-m')
  if cpu_index >= 0:
    resource_dict['cpu'] = cmd_list[cpu_index + 1].split(',')[0]
  if ram_index >= 0:
    resource_dict['ram'] = cmd_list[ram_index + 1].split(',')[0]

  return resource_dict

class QmpCommandError(Exception):
  pass

class QmpDeviceRemoveError(Exception):
  pass

class QemuQMPWrapper(object):
  """
  Small wrapper around Qemu's QMP to control a qemu VM.
  See http://git.qemu.org/?p=qemu.git;a=blob;f=qmp-commands.hx for
  QMP API definition.
  """
  def __init__(self, unix_socket_location, auto_connect=True):
    self._event_list = []
    self.socket = None
    self.sockf = None
    if auto_connect:
      self.socket = self.connectToQemu(unix_socket_location)
      self.sockf = self.socket.makefile()
      self.capabilities()

  @staticmethod
  def connectToQemu(unix_socket_location):
    """
    Create a socket, connect to qemu, be sure it answers, return socket.
    """
    if not os.path.exists(unix_socket_location):
      raise Exception('unix socket %s does not exist.' % unix_socket_location)

    print('Connecting to qemu...')
    so = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    connected = False
    while not connected:
      try:
        so.connect(unix_socket_location)
      except socket.error:
        time.sleep(1)
        print('Could not connect, retrying...')
      else:
        connected = True
    so.recv(1024)

    return so

  def _readResponse(self, only_event=False):
    response = None
    while True:
      data = self.sockf.readline()
      if not data:
        return
      try:
        response = json.loads(data)
      except ValueError:
        raise ValueError('Wrong data: %s' % data)
      if 'error' in response:
        raise QmpCommandError(response["error"]["desc"])
      if 'event' in response:
        self._event_list.append(response)
        print(response)
        if not only_event:
          continue

      return response

  def _send(self, message, retry=0, sleep=0.5):
    for i in range(retry + 1):
      self.socket.sendall(str2bytes(json.dumps(message)))
      response = self._readResponse()
      if response is not None:
        break
      if i < retry:
        print("Retrying send command after %s second(s)..." % sleep)
        time.sleep(sleep)
    return response

  def _getVMStatus(self):
    response = self._send({'execute': 'query-status'})
    if response:
      return self._send({'execute': 'query-status'})['return']['status']
    else:
      raise IOError('Empty answer')

  def _waitForVMStatus(self, wanted_status):
    while True:
      try:
        actual_status = self._getVMStatus()
        if actual_status == wanted_status:
          return
        else:
          print('VM in %s status, wanting it to be %s, retrying...' % (
              actual_status, wanted_status))
          time.sleep(1)
      except IOError:
          print('VM not ready, retrying...')

  def capabilities(self):
    print('Asking for capabilities...')
    self._send({'execute': 'qmp_capabilities'})

  def getEventList(self, timeout=0, cleanup=False):
    """
    Get a list of available QMP events.
    """

    if self.socket is None:
      return []
    if cleanup:
      self.cleanupEventList()
    if not self._event_list and timeout > 0:
      self.socket.settimeout(timeout)
      try:
        # Read then wait a bit
        self._readResponse(only_event=True)
      except socket.timeout:
        pass
      finally:
        self.socket.settimeout(None)
    else:
      self.socket.setblocking(0)
      try:
        self._readResponse(only_event=True)
      except socket.error as err:
        if err[0] == errno.EAGAIN:
          # No data available
          pass
      finally:
        self.socket.setblocking(1)

    return self._event_list

  def cleanupEventList(self):
    """
    Clear current list of pending events.
    """
    self.__events = []

  def setVNCPassword(self, password):
    # Set VNC password
    print('Setting VNC password...')
    result = self._send({
      "execute": "change",
      "arguments": {
        "device": "vnc",
        "target": "password",
        "arg": password
      }
    })
    if result and result.get('return', None) != {}:
      raise ValueError(result)
    print('Done.')

  def powerdown(self):
    print('Stopping the VM...')
    self._send({'execute': 'system_powerdown'})

  def suspend(self):
    print('Suspending VM...')
    self._send({'execute': 'stop'})
    self._waitForVMStatus('paused')

  def resume(self):
    print('Resuming VM...')
    self._send({'execute': 'cont'})
    self._waitForVMStatus('running')

  def _queryBlockJobs(self, device):
    return self._send({'execute': 'query-block-jobs'})

  def _getRunningJobList(self, device):
    result = self._queryBlockJobs(device)
    max_loop = 20

    while max_loop and result is None:
      # If result is None retry to command until if return something.
      max_loop -= 1
      time.sleep(3)
      result = self._queryBlockJobs(device)
    if result and result.get('return'):
      return result['return']
    else:
      return

  def driveBackup(self, backup_target, source_device='virtio0', sync_type='full'):
    print('Asking Qemu to perform backup to %s' % backup_target)
    # XXX: check for error
    self._send({
        'execute': 'drive-backup',
        'arguments': {
            'device': source_device,
            'sync': sync_type,
            'target': backup_target,
         }
    })
    while self._getRunningJobList(backup_target):
      print('Job is not finished yet.')
      time.sleep(20)

  def createSnapshot(self, snapshot_file, device='virtio0'):
    print(self._send({
        'execute': 'blockdev-snapshot-sync',
        'arguments': {
            'device': device,
            'snapshot-file': snapshot_file,
         }
    }))

  def createInternalSnapshot(self, name=None, device='virtio0'):
    if name is None:
      name = int(time.time())
    self._send({
        'execute': 'blockdev-snapshot-internal-sync',
        'arguments': {
            'device': device,
            'name': name,
         }
    })

  def deleteInternalSnapshot(self, name, device='virtio0'):
    self._send({
        'execute': 'blockdev-snapshot-delete-internal-sync',
        'arguments': {
            'device': device,
            'name': name,
         }
    })

  def getCPUInfo(self):
    """
      return some info about VM CPUs
    """
    cpu_info_dict = {'hotplugged': [], 'base': []}
    cpu_list = self._send({
      'execute': 'query-cpus'
    }, retry=5)['return']
    for cpu in cpu_list:
      if 'unattached' in cpu['qom_path']:
        index = 'base'
      else:
        index = 'hotplugged'
      cpu_info_dict[index].append({
        'props': cpu['props'],
        'CPU': cpu['CPU'],
        'qom_path': cpu['qom_path']
      })
    return cpu_info_dict

  def getMemoryInfo(self):
    """
      return some info about VM Memory. Can only say info about hotplugged RAM
    """
    mem_info_dict = {'hotplugged': [], 'base': []}
    memory_list = self._send({
      'execute': 'query-memory-devices'
    }, retry=5)['return']
    for mem in memory_list:
      if mem['data']['hotplugged'] == True:
        mem_info_dict['hotplugged'].append(mem['data'])

    return mem_info_dict

  def _removeDevice(self, dev_id, command_dict, auto_reboot=False):
    max_retry = 5
    result = None
    resend = True
    stop_retry = False
    while max_retry > 0:
      max_retry -= 1
      try:
        if resend:
          result = self._send(command_dict)
      except QmpCommandError as e:
        print("ERROR: ", e)
        print("%s\nRetry remove %r in few seconds..." % (result, dev_id))
        resend = True
      else:
        for event in self.getEventList(timeout=2, cleanup=True):
          if 'ACPI_DEVICE_OST' == event['event']:
            # request was received
            resend = False
          if 'DEVICE_DELETED' == event['event']:
            # device was deleted
            stop_retry = True
            break
        if stop_retry:
          break
        elif result is None and max_retry > 0:
          print("Retry remove %r in few seconds..." % dev_id)
      time.sleep(2)

    if result is not None:
      if result.get('return', None) == {} or ('error' in result and \
          result['error'].get('class', '') == 'DeviceNotFound'):
        print('Device %s was removed.' % dev_id)
        return

    # device was not remove after retries
    if not auto_reboot:
      raise ValueError("Cannot remove device %s" % dev_id)

    # try soft reboot of the VM
    self.powerdown()
    system_exited = False
    # wait for ~10 seconds if the system exit, else quit Qemu
    for i in range(0, 5):
      for event in self.getEventList(timeout=2, cleanup=True):
        if event['event'] == 'SHUTDOWN':
          # qemu is about to exit
          system_exited = True
      if system_exited:
        break
      else:
        time.sleep(2)

    if not system_exited:
      # hard reset the VM
      print("Trying hard shutdown of the VM...")
      self._send({"execute": "quit"})

    raise QmpDeviceRemoveError("Stopped Qemu in order to remove the device %r" % dev_id)


  def _updateCPU(self, amount, cpu_model):
    """
      Add or remove CPU according current value
      amount: number of CPU to update to
    """
    cpu_amount = 0
    empty_socket_list = []
    used_socket_id_list = []
    unremovable_cpu = 0
    cpu_hotplugable_list = self._send({
      'execute': 'query-hotpluggable-cpus'
    }, retry=5)['return']
    cpu_hotplugable_list.reverse()
    for cpu in cpu_hotplugable_list:
      if cpu.get('qom-path', '') == '':
        if len(empty_socket_list) < amount:
          cpu['props']['driver'] = cpu_model
          cpu['props']['id'] = 'cpu%s' % (cpu['props']['socket-id'])
          empty_socket_list.append(cpu['props'])
      else:
        # if this is an hotpluggable cpu
        if '/machine/peripheral' in cpu.get('qom-path', ''):
          used_socket_id_list.append('cpu%s' % (cpu['props']['socket-id']))
          cpu_amount += 1
        else:
          unremovable_cpu += 1

    hotplug_amount = amount - unremovable_cpu # get only hotpluggable CPU
    if hotplug_amount < 0:
      raise ValueError("Unattached CPU amount is %s, cannot update to %s" % (
        unremovable_cpu, amount))
    cpu_diff = hotplug_amount - cpu_amount
    max_hotplug_cpu = len(empty_socket_list)

    if cpu_amount == hotplug_amount:
      # no chanches
      print("Hotplug CPU is up to date.")
      return

    if cpu_amount > hotplug_amount:
      # we will remove CPU
      cpu_diff = -1 * cpu_diff
      if cpu_diff >= 1:
        print("Request remove %s CPUs..." % cpu_diff)
        used_socket_id_list.reverse()
        for i in range(0, cpu_diff):
          self._removeDevice(used_socket_id_list[i], {
            'execute': 'device_del',
            'arguments': {'id': used_socket_id_list[i]}
          })
    elif cpu_amount < hotplug_amount:
      if max_hotplug_cpu < cpu_diff:
        # no hotplugable cpu socket found for Add
        raise ValueError("Cannot Configure %s CPUs, the maximum amount of " \
          "hotplugable CPU is %s!" % (hotplug_amount, max_hotplug_cpu))
      print("Adding %s CPUs..." % cpu_diff)
      for i in range(0, cpu_diff):
        self._send({
            'execute': 'device_add',
            'arguments': empty_socket_list[i]
        })

    # check that hotplugged memery amount is consistent
    cpu_info = self.getCPUInfo()
    final_cpu_count = len(cpu_info['hotplugged'])
    if hotplug_amount != final_cpu_count:
      raise ValueError("Consistency error: Expected %s hotplugged CPU(s) but" \
        " current CPU amount is %s" % (hotplug_amount, final_cpu_count))
    print("Done.")

  def _removeMemory(self, id_dict, auto_reboot=False):
    print("Trying to remove devices %s, %s..." % (id_dict['id'], id_dict['memdev']))
    self._removeDevice(id_dict['id'] ,{
      'execute': 'device_del',
      'arguments': {'id': id_dict['id']}
    }, auto_reboot=auto_reboot)
    # when dimm is removed, remove memdev object
    self._removeDevice(id_dict['memdev'], {
      'execute': 'object-del',
      'arguments': {
        'id': id_dict['memdev']
      }
    }, auto_reboot=auto_reboot)

  def _updateMemory(self, mem_size, slot_size, slot_amount, allow_reboot=False):
    """
      Update memory size according to the current value. option_dict contains:

      slot_amount: amount of slots available
      mem_size: Size of memory to allocate
      slot_size: size of the memory per slot (this value should not change).
        default: 512 MB

      ex: to add 2G of RAM, mem=2048,slot=512 => allocate 4 slots
          to reduce to 1G of RAM, mem=1024,slot=512 => allocate 2 slots
    """

    current_size = 0
    num_slot_used = 0
    memory_id_list = [] # current hotplugged memory
    cleanup_memdev_id_dict = {}
    current_dimm_list = self._send({ "execute": "query-memory-devices" }, retry=5)
    current_memdev_list = self._send({ "execute": "query-memdev" }, retry=5)

    for memdev in current_memdev_list['return']:
      # qemu 5.x+ reports init memory static memory as pc.ram
      if memdev['id'] != 'pc.ram':
        cleanup_memdev_id_dict[memdev['id']] = ''

    for dimm in current_dimm_list['return']:
      current_size += dimm['data']['size']
      if dimm['data']['hotplugged']:
        mem_dev = os.path.basename(dimm['data']['memdev'])
        cleanup_memdev_id_dict.pop(mem_dev)
        memory_id_list.append({
        'memdev': mem_dev,
         'id': dimm['data']['id'],
         'size': dimm['data']['size']//(1024 * 1024),
       })
    memory_id_list = sorted(memory_id_list, key=itemgetter('id'))

    # cleanup memdev that was not removed because of failure
    for memdev in cleanup_memdev_id_dict.keys():
      print("Cleaning up memdev %s..." % memdev)
      self._removeDevice(memdev, {
        'execute': 'object-del',
        'arguments': {
          'id': memdev
        }
      }, auto_reboot=allow_reboot)

    num_slot_used = len(memory_id_list)
    if (mem_size % slot_size) != 0:
      raise ValueError("Memory size %r is not a multiple of %r" % (mem_size,
        slot_size))
    if (mem_size // slot_size) > slot_amount:
      raise ValueError("No enough slots available to add %sMB of RAM" % mem_size)

    current_size //= (1024 * 1024)
    if current_size == mem_size:
      print("Hotplug Memory size is up to date.")
      return

    if mem_size < 0:
      raise ValueError("Memory size is not valid: %s" % option_dict)
    elif current_size > mem_size:
      # Request to remove memory
      to_remove_size = current_size - mem_size
      print("Removing %s MB of memory..." % to_remove_size)

      for i in range(num_slot_used, 0, -1):
        # remove all slots that won't be used
        index = i - 1
        to_remove_size -= memory_id_list[index]['size']
        if to_remove_size >= 0:
          self._removeMemory(memory_id_list[index], auto_reboot=allow_reboot)
          if to_remove_size == 0:
            break
        else:
          raise ValueError("Cannot remove the requested size of memory. " \
            "Remaining to remove: %s, RAM slot size: %s" % (
              to_remove_size + memory_id_list[index]['size'],
              memory_id_list[index]['size'])
            )
    elif current_size < mem_size:
      # ask for increase memory
      slot_add = (mem_size - current_size) // slot_size

      print("Adding %s memory slot(s) of %s MB..." % (slot_add, slot_size))
      for i in range(0, slot_add):
        index = num_slot_used + i + 1
        self._send({
            'execute': 'object-add',
            'arguments': {
              'qom-type': 'memory-backend-ram',
              'id': 'mem%s' % index,
              'props': { 'size': slot_size * 1024 * 1024 }
            }
        })
        self._send({
            'execute': 'device_add',
            'arguments': {
              'driver': 'pc-dimm' ,
              'id': 'dimm%s' % index,
              'memdev': 'mem%s' % index
            }
          })

    # check that hotplugged memery amount is consistent
    mem_info = self.getMemoryInfo()
    final_mem_size = 0
    for mem in mem_info['hotplugged']:
      final_mem_size += mem['size']
    final_mem_size //= (1024 * 1024) # get size in MB
    if mem_size != final_mem_size:
      raise ValueError("Consistency error: Expected %s MB of hotplugged RAM " \
        "but current RAM size is %s MB" % (mem_size, final_mem_size))
    print("Done.")

  def updateDevice(self, option_dict):
    argument_dict = {}
    if 'device' in option_dict:
      if option_dict['device'] == 'cpu':
        return self._updateCPU(
          amount=int(option_dict['amount']),
          cpu_model=option_dict.get('model', 'qemu64-x86_64-cpu')
        )
      elif option_dict['device'] == 'memory':
        return self._updateMemory(
          mem_size=int(option_dict['mem']),
          slot_size=int(option_dict['slot']),
          slot_amount=int(option_dict['nslot']),
          allow_reboot=int(option_dict.get('canreboot', 0)) in [1, True]
        )
      else:
        raise ValueError("Unknown device type: %s" % option_dict)
    else:
      raise ValueError("Options are unknown: %s" % option_dict)

  def queryCommands(self, query=None):
    if query is not None:
      pprint.pprint(self._send({'execute': 'query-%s' % query})['return'])
    else:
      pprint.pprint(self._send({'execute': 'query-commands'})['return'])

def main():
  config = parseArgument()
  qemu_wrapper = QemuQMPWrapper(config.unix_socket_location)

  if config.remainding_argument_list:
    getattr(qemu_wrapper, config.action)(*config.remainding_argument_list)
  else:
    if config.query != None:
      getattr(qemu_wrapper, config.action)(**{"query": config.query})
    elif config.action =="updateDevice":
      argument_dict = {}
      for parameter in config.device_options.split(','):
        parameter_list = parameter.split('=')
        argument_dict[parameter_list[0].strip()] = parameter_list[1].strip()
      getattr(qemu_wrapper, config.action)(argument_dict)
    else:
      getattr(qemu_wrapper, config.action)()

if __name__ == '__main__':
  main()




import json
import multiprocessing
import os
import socket
import time
import unittest

from slapos.equeue import EqueueServer


class Options:
  def __init__(self, logfile, loglevel, database, takeover_triggered_file_path, lockfile, timeout):
    self.logfile = logfile
    self.loglevel = loglevel
    self.database = database
    self.takeover_triggered_file_path = takeover_triggered_file_path
    self.lockfile = lockfile
    self.timeout = timeout


class TestEqueue(unittest.TestCase):
  def setUp(self):
    cwd = os.getcwd()
    (
      self.database,
      self.takeover_triggered_file_path,
      self.lockfile,
      self.socket,
      self.testfile,
    ) = self.file_list = [
      'equeue.db',
      'takeover.txt',
      'equeue.lock',
      'equeue.sock',
      'testfile.txt',
    ]
    # Do not delete log files, as it may contain interesting information for debugging
    self.logfile = 'equeue.log'
    self.options = Options(
      logfile=self.logfile,
      loglevel='INFO',
      database=self.database,
      takeover_triggered_file_path=self.takeover_triggered_file_path,
      lockfile=self.lockfile,
      timeout=3,
    )
    self.equeue_server = None


  def tearDown(self):
    self.equeue_server.server_close()
    for filename in self.file_list:
      if os.path.exists(filename):
        os.unlink(filename)


  def test_onlyOneCommandRunAtTheTime(self):
    self.equeue_server = EqueueServer(self.socket, equeue_options=self.options)
    server_process = multiprocessing.Process(target=self.equeue_server.serve_forever)
    server_process.start()
    self.addCleanup(server_process.terminate)

    self.assertTrue(os.path.exists(self.options.database))
    self.assertTrue(os.path.exists(self.socket))

    # Prepare 2 requests, the first one including a delay.
    # If more than one request can be run at a time, then
    # the second would execute first, and the content of the
    # test file would be overwriten by command 1.
    request_1 = {
      'timestamp': 1,
      'command': "bash\0-c\0sleep 3 && echo 1 > %s" % self.testfile,
    }
    request_2 = {
      'timestamp': 2,
      'command': "bash\0-c\0echo 2 > %s" % self.testfile,
    }
    for request in (request_1, request_2):
      equeue_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      equeue_socket.connect(self.socket)
      equeue_socket.send(json.dumps(request).encode())
      result = equeue_socket.recv(len(request['command']))
      self.assertEqual(result.decode(), request['command'])
      equeue_socket.close()

    # Wait for the 1st command to finish
    time.sleep(5)

    with open(self.testfile, 'r') as fd:
      self.assertEqual(fd.read(), '2\n')


  def test_doNotRunACommandIfItHasAlreadyRun(self):
    self.equeue_server = EqueueServer(self.socket, equeue_options=self.options)
    server_process = multiprocessing.Process(target=self.equeue_server.serve_forever)
    server_process.start()
    self.addCleanup(server_process.terminate)

    self.assertTrue(os.path.exists(self.options.database))
    self.assertTrue(os.path.exists(self.socket))

    # Run 2 times the same command (bash), but with different
    # parameters (equeue doesn't care about parameters). As the
    # timestamp is the same, the second command shouldn't be run
    request_1 = {
      'timestamp': 1,
      'command': "bash\0-c\0echo 1 > %s" % self.testfile,
    }
    request_2 = {
      'timestamp': 1,
      'command': "bash\0-c\0echo 2 > %s" % self.testfile,
    }
    for request in (request_1, request_2):
      equeue_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      equeue_socket.connect(self.socket)
      equeue_socket.send(json.dumps(request).encode())
      result = equeue_socket.recv(len(request['command']))
      self.assertEqual(result.decode(), request['command'])
      equeue_socket.close()

    # Give a bit of time for command to run
    time.sleep(2)

    with open(self.testfile, 'r') as fd:
      self.assertEqual(fd.read(), '1\n')


  def test_doNothingIfTakeoverHasBeenTriggered(self):
    self.equeue_server = EqueueServer(self.socket, equeue_options=self.options)
    server_process = multiprocessing.Process(target=self.equeue_server.serve_forever)
    server_process.start()
    self.addCleanup(server_process.terminate)

    self.assertTrue(os.path.exists(self.options.database))
    self.assertTrue(os.path.exists(self.socket))

    # Create takeover file
    with open(self.takeover_triggered_file_path, 'w') as f:
      f.write('\0')

    request = {
      'timestamp': 1,
      'command': "bash\0-c\0echo 1 > %s" % self.testfile,
    }
    equeue_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    equeue_socket.connect(self.socket)
    equeue_socket.send(json.dumps(request).encode())
    result = equeue_socket.recv(len(request['command']))
    self.assertEqual(result.decode(), request['command'])
    equeue_socket.close()

    # Give a bit of time for command to run
    time.sleep(2)

    self.assertFalse(os.path.exists(self.testfile))



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

from slapos.grid.svcbackend import getSupervisorRPC
from slapos.grid.svcbackend import _getSupervisordSocketPath

import os


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # SR can set custom periodicity
    self.setPeriodicity(float(self.getConfig('frequency', 2)))
    self.result_count = int(self.getConfig('result-count', 3))
    self.failure_amount = int(self.getConfig('failure-amount', 3))

  def sense(self):
    """
      Check if a service is in the expected state
    """

    service = self.getConfig('service')
    expect = self.getConfig('expect', '')

    if expect not in ("running", "stopped"):
      self.logger.info("OK service %r is allowed to be in any state (expect = %r)", service, expect)
      return

    if not service:
      self.logger.error("ERROR no service is specified")
      return

    run_directory = self.getConfig('run-directory', os.path.join(self.getPartitionFolder(), "srv", "runner", "var", "run"))

    try:
      supervisor_rpc = getSupervisorRPC(_getSupervisordSocketPath(run_directory, self.logger))

      with supervisor_rpc as supervisor:
        state = supervisor.getProcessInfo(service)['statename'].lower()
        if state != expect:
          self.logger.error("ERROR service %r is in state %r (expected %r)", service, state, expect)
        else:
          self.logger.info("OK service %r is in expected state %r", service, state)

    except Exception as e:
      self.logger.error("ERROR %r", e, exc_info=True)

  def anomaly(self):
    return self._anomaly(result_count=self.result_count, failure_amount=self.failure_amount)



from __future__ import print_function

import argparse
import subprocess

from .runner_utils import *

def parsePostNotificationRunArgumentList():
  parser = argparse.ArgumentParser()
  parser.add_argument('--srv-path', required=True)
  parser.add_argument('--backup-path', required=True)
  parser.add_argument('--diff-file', required=True)
  parser.add_argument('--proof-signature-file', required=True)

  return parser.parse_args()

def postNotificationRun():
  args = parsePostNotificationRunArgumentList()

  runner_working_path = os.path.join(args.srv_path, 'runner')

  with CwdContextManager(runner_working_path):
    slappart_signature_method_dict = getSlappartSignatureMethodDict()

  with CwdContextManager(args.backup_path):
    writeSignatureFile(slappart_signature_method_dict, runner_working_path, args.proof_signature_file)

    try:
      diff_output = subprocess.check_output(['diff',  '-ruw', 'backup.signature', args.proof_signature_file])
    except subprocess.CalledProcessError as e:
      diff_output = e.output
    finally:
      with open(args.diff_file, 'w') as f:
        f.write(diff_output)
